"""
WiP.

Soon.
"""

# region [Imports]

import os
import re
import sys
import json
import queue
import math
import base64
import pickle
import random
import shelve
import dataclasses
import shutil
import asyncio
import logging
import sqlite3
import platform

import subprocess
import inspect

from time import sleep, process_time, process_time_ns, perf_counter, perf_counter_ns
from io import BytesIO, StringIO
from abc import ABC, ABCMeta, abstractmethod
from copy import copy, deepcopy
from enum import Enum, Flag, auto, unique
from pprint import pprint, pformat
from pathlib import Path
from string import Formatter, digits, printable, whitespace, punctuation, ascii_letters, ascii_lowercase, ascii_uppercase
from timeit import Timer
from typing import (TYPE_CHECKING, TypeVar, TypeGuard, TypeAlias, Final, TypedDict, Generic, Union, Optional, ForwardRef, final,
                    no_type_check, no_type_check_decorator, overload, get_type_hints, cast, Protocol, runtime_checkable, NoReturn, NewType, Literal, AnyStr, IO, BinaryIO, TextIO, Any)
from collections import Counter, ChainMap, deque, namedtuple, defaultdict
from collections.abc import (AsyncGenerator, AsyncIterable, AsyncIterator, Awaitable, ByteString, Callable, Collection, Container, Coroutine, Generator,
                             Hashable, ItemsView, Iterable, Iterator, KeysView, Mapping, MappingView, MutableMapping, MutableSequence, MutableSet, Reversible, Sequence, Set, Sized, ValuesView)
from zipfile import ZipFile, ZIP_LZMA
from datetime import datetime, timezone, timedelta
from tempfile import TemporaryDirectory
from textwrap import TextWrapper, fill, wrap, dedent, indent, shorten
from functools import wraps, partial, lru_cache, singledispatch, total_ordering, cached_property, cache
from contextlib import contextmanager, asynccontextmanager, nullcontext, closing, ExitStack, suppress
from statistics import mean, mode, stdev, median, variance, pvariance, harmonic_mean, median_grouped
from urllib.parse import urlparse
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, wait, as_completed, ALL_COMPLETED, FIRST_EXCEPTION, FIRST_COMPLETED
from fontTools import ttLib


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    ...


# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


WINDOWS_FONT_FOLDER: tuple[Path] = (Path(os.getenv("WINDIR"), "Fonts"),
                                    Path(os.getenv("LOCALAPPDATA"), "Microsoft", "Windows", "Fonts"))


class FontFile:
    __slots__ = ("_file_path",
                 "_name",
                 "_family_name")

    def __init__(self, file_path: Union[str, os.PathLike, Path]) -> None:
        self._file_path = Path(file_path).resolve()
        self._name: str = None
        self._family_name: str = None

    @property
    def name(self) -> str:
        if self._name is None:
            self._name, self._family_name = self.get_name_and_family_name(self.get_font_obj())
        return self._name

    @property
    def family_name(self) -> str:
        if self._family_name is None:
            self._name, self._family_name = self.get_name_and_family_name(self.get_font_obj())
        return self._family_name

    @property
    def file_path(self) -> Path:
        return self._file_path

    def get_font_obj(self) -> ttLib.TTFont:
        return ttLib.TTFont(self.file_path)

    def __fspath__(self) -> str:
        return os.fspath(self.file_path)

    def __hash__(self) -> int:
        return hash(self.name) + hash(self.family_name)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name={self.name!r}, family={self.family_name!r})'

    @staticmethod
    def get_name_and_family_name(font_item) -> tuple[str, str]:
        """Get the short name from the font's names table"""
        FONT_SPECIFIER_NAME_ID = 4
        FONT_SPECIFIER_FAMILY_ID = 1

        name = ""
        family = ""
        for record in font_item['name'].names:
            if b'\x00' in record.string:
                name_str = record.string.decode('utf-16-be', errors="ignore")
            else:
                name_str = record.string.decode('utf-8', errors="ignore")
            if record.nameID == FONT_SPECIFIER_NAME_ID and not name:
                name = name_str
            elif record.nameID == FONT_SPECIFIER_FAMILY_ID and not family:
                family = name_str
            if name and family:
                break
        return name, family


def iter_all_fonts() -> Generator["FontFile", None, None]:
    for font_folder in WINDOWS_FONT_FOLDER:

        for file in font_folder.iterdir():
            if file.is_dir():
                continue

            if file.suffix != ".ttf":
                continue

            yield FontFile(file)


def list_all_fonts() -> list["FontFile"]:
    return list(iter_all_fonts())


class _Font_Cache:

    def __init__(self) -> None:
        self._all_fonts: set["FontFile"] = None
        self._name_data: dict[str, "FontFile"] = {}
        self._family_and_name_data: dict[tuple[str, str], "FontFile"] = {}
        self._file_path_data: dict[Path, "FontFile"] = {}

    @property
    def all_fonts(self) -> set["FontFile"]:
        if self._all_fonts is None:
            self._all_fonts = set(iter_all_fonts())
        return self._all_fonts

    def _load_by_name(self, name: str) -> Optional["FontFile"]:
        return next((font for font in self.all_fonts if font.name == name), None)

    def _load_by_family_and_name(self, name: str, family_name: str) -> Optional["FontFile"]:
        return next((font for font in self.all_fonts if font.family_name == family_name and font.name == name), None)

    def _load_by_path(self, path: Path) -> Optional["FontFile"]:
        if path.exists() is True:
            return FontFile(path)

    def store_font(self, font: "FontFile") -> None:
        self._name_data[font.name] = font
        self._family_and_name_data[(font.family_name, font.name)] = font
        self._file_path_data[font.file_path] = font

    def __getitem__(self, key: Union[str, tuple[str, str], Path]) -> "FontFile":
        if isinstance(key, Path):
            try:
                font = self._file_path_data[key]
            except KeyError as e:
                font = self._load_by_path(key)
                if font is None:
                    raise e
                self.store_font(font)

        elif isinstance(key, tuple):
            try:
                font = self._family_and_name_data[key]
            except KeyError as e:
                font = self._load_by_family_and_name(family_name=key[0], name=key[1])
                if font is None:
                    raise e
                self.store_font(font)

        elif isinstance(key, str):
            try:
                font = self._name_data[key]
            except KeyError as e:
                font = self._load_by_name(key)
                if font is None:
                    raise e
                self.store_font(font)

        return font


FONT_STORAGE = _Font_Cache()

# region [Main_Exec]


if __name__ == '__main__':
    from pympler.asizeof import asizeof
    from gidapptools.general_helper.conversion import bytes2human

    # for i in iter_all_fonts():
    #     print(i)

    print(FONT_STORAGE["Viking Squad Halftone"])
    print(FONT_STORAGE["Viking Squad Halftone"])

    print(FONT_STORAGE[('Source Sans Pro Black', 'Source Sans Pro Black')])

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

import os
import re
import sys
import json
import queue
import math
import base64
import pickle
import random
import shelve
import dataclasses
import shutil
import asyncio
import logging
import sqlite3
import platform

import subprocess
import inspect

from time import sleep, process_time, process_time_ns, perf_counter, perf_counter_ns
from io import BytesIO, StringIO
from abc import ABC, ABCMeta, abstractmethod
from copy import copy, deepcopy
from enum import Enum, Flag, auto, unique
from pprint import pprint, pformat
from pathlib import Path
from string import Formatter, digits, printable, whitespace, punctuation, ascii_letters, ascii_lowercase, ascii_uppercase
from timeit import Timer
from typing import (TYPE_CHECKING, TypeVar, TypeGuard, TypeAlias, Final, TypedDict, Generic, Union, Optional, ForwardRef, final, Callable,
                    no_type_check, no_type_check_decorator, NamedTuple, overload, get_type_hints, cast, Protocol, runtime_checkable, NoReturn, NewType, Literal, AnyStr, IO, BinaryIO, TextIO, Any)
from collections import Counter, ChainMap, deque, defaultdict
from collections.abc import (AsyncGenerator, AsyncIterable, AsyncIterator, Awaitable, ByteString, Callable, Collection, Container, Coroutine, Generator,
                             Hashable, ItemsView, Iterable, Iterator, KeysView, Mapping, MappingView, MutableMapping, MutableSequence, MutableSet, Reversible, Sequence, Set, Sized, ValuesView)
from zipfile import ZipFile, ZIP_LZMA
from datetime import datetime, timezone, timedelta
from tempfile import TemporaryDirectory
from textwrap import TextWrapper, fill, wrap, dedent, indent, shorten
from functools import wraps, partial, lru_cache, singledispatch, total_ordering, cached_property, cache
from contextlib import contextmanager, asynccontextmanager, nullcontext, closing, ExitStack, suppress
from statistics import mean, mode, stdev, median, variance, pvariance, harmonic_mean, median_grouped
from urllib.parse import urlparse
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, wait, as_completed, ALL_COMPLETED, FIRST_EXCEPTION, FIRST_COMPLETED


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    ...

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def get_parent_class(in_class: type) -> Union[type, None]:
    try:
        return in_class.mro()[1]
    except IndexError:
        return None


def iter_all_child_classes(start_class: type) -> Generator[type, None, None]:
    for child_class in start_class.__subclasses__():
        yield child_class
        yield from iter_all_child_classes(child_class)


CLASS_TREE_NODE_KLASS_TYPE = TypeVar("CLASS_TREE_NODE_KLASS_TYPE", bound=type)


class ClassTreeNode(Generic[CLASS_TREE_NODE_KLASS_TYPE]):

    def __init__(self, klass: CLASS_TREE_NODE_KLASS_TYPE) -> None:
        self._klass = klass
        self._parent_node: Union["ClassTreeNode", None] = None
        self._child_nodes: tuple["ClassTreeNode"] = tuple()

    @property
    def klass(self) -> CLASS_TREE_NODE_KLASS_TYPE:
        return self._klass

    @property
    def parent_node(self) -> Union["ClassTreeNode", None]:
        return self._parent_node

    @property
    def child_nodes(self) -> tuple["ClassTreeNode"]:
        return self._child_nodes

    def set_parent_node(self, parent_node: "ClassTreeNode") -> None:
        self._parent_node = parent_node

    def add_child_node(self, child_node: "ClassTreeNode") -> None:
        self._child_nodes = self._child_nodes + (child_node,)

    def get_child_by_klass(self, klass: CLASS_TREE_NODE_KLASS_TYPE) -> "ClassTreeNode":
        result = next((child for child in self.child_nodes if child.klass is klass), None)
        if result is None:
            raise KeyError(f"{self!r} has no child_node for klass{klass!r}.")

        return result

    def get_child_by_klass_recursively(self, klass: CLASS_TREE_NODE_KLASS_TYPE) -> Union["ClassTreeNode", None]:

        def _iter_child_node_recursive(start_node: "ClassTreeNode") -> Generator["ClassTreeNode", None, None]:
            for _child_node in start_node.child_nodes:
                yield _child_node
                yield from _iter_child_node_recursive(_child_node)
        if klass is self.klass:
            return self
        for child_node in _iter_child_node_recursive(self):
            if child_node.klass is klass:
                return child_node

        raise KeyError(f"{self!r} has no child_node recursively for klass{klass!r}.")

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(klass={self.klass!r})'


class ClassTree:

    def __init__(self, start_class: CLASS_TREE_NODE_KLASS_TYPE) -> None:
        self.root: ClassTreeNode = ClassTreeNode(start_class)

    def _fill(self) -> None:
        for child_class in iter_all_child_classes(self.root.klass):

            parent_node = self.root.get_child_by_klass_recursively(get_parent_class(child_class))

            if parent_node is None:
                print("--------------------")
                print(f"{child_class=}")
                print(f"{get_parent_class(child_class)=}")
                print(f"{self.root.child_nodes=}")
                print("--------------------")
                # parent_node = self.root
            parent_node.add_child_node(ClassTreeNode(child_class))

            print("=============================")

    def draw(self, engine: str = "ASCII") -> None:
        print(f"{self.root.klass.__name__}")

        for child_node in self.root.child_nodes:
            print(f"    ├─{child_node.klass.__name__}")
            for sub_child_node in child_node.child_nodes:
                print(f"    │     ├─{sub_child_node.klass.__name__}")
            print("    │")


        # region [Main_Exec]
if __name__ == '__main__':
    import pp
    from gidapptools.errors import GidAppToolsBaseError
    xxx = ClassTree(GidAppToolsBaseError)
    xxx._fill()
    xxx.draw()

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from typing import ClassVar, Optional, Union
from pathlib import Path
import json
import sys
import re
# * Third Party Imports --------------------------------------------------------------------------------->
import attrs
from threading import Lock
import psutil
from math import ceil, floor
# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import ApplicationInstanceAlreadyRunningError, show_basic_error_pop_up
from gidapptools.custom_types import PATH_TYPE
from gidapptools.general_helper.path_helper import ensure_valid_file_stem
import sys
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


@attrs.define(frozen=True, slots=True)
class LockFileData:
    text_separator: ClassVar[str] = "|:|"
    app_name: str = attrs.field()
    pid: int = attrs.field()
    creation_time: float = attrs.field()

    @creation_time.default
    def _default_creation_time(self) -> float:
        return floor(self.get_process().create_time())

    def get_process(self) -> Union[None, psutil.Process]:
        try:
            process = psutil.Process(self.pid)

            if process.is_running() is False:
                return None

            return process

        except psutil.NoSuchProcess:
            return None

    @property
    def is_running(self) -> bool:
        process = self.get_process()

        if process is None:
            return False

        if floor(process.create_time()) != self.creation_time:
            return False

        return True

    @classmethod
    def from_file(cls, file_path: PATH_TYPE) -> "LockFileData":
        content = Path(file_path).resolve().read_text(encoding='utf-8', errors='ignore').strip()
        app_name, raw_pid, raw_creation_time = (p.strip() for p in content.split(cls.text_separator))
        pid = int(raw_pid)
        creation_time = floor(float(raw_creation_time))
        return cls(app_name=app_name, pid=pid, creation_time=creation_time)

    def to_text(self) -> str:
        return f"{self.app_name}{self.text_separator}{self.pid}{self.text_separator}{self.creation_time}"


_SAFETY_LOCK = Lock()


class SingleRunningInstanceRestrictor:
    _lock_file_suffix: str = ".running_instance"

    def __init__(self, storage_folder: PATH_TYPE, app_name: str) -> None:
        self._storage_folder = Path(storage_folder).resolve()
        self._app_name = app_name
        self._pid = os.getpid()
        self._current_process_lock_file_data = LockFileData(app_name=self._app_name, pid=self._pid)
        self._lock_file_path: Path = (self._storage_folder / (ensure_valid_file_stem(self._app_name, True, True) + self._lock_file_suffix)).resolve()
        self._lock_succeded: bool = False

    @property
    def lock_file_name(self) -> str:
        return self._lock_file_path.name

    @property
    def lock_file_path(self) -> Path:
        return self._lock_file_path

    @property
    def lock_file_exists(self) -> bool:
        return self.lock_file_path.is_file()

    def get_existing_lock_file_data(self) -> LockFileData:
        return LockFileData.from_file(self.lock_file_path)

    def store_in_lock_file(self) -> None:
        self.lock_file_path.parent.mkdir(exist_ok=True, parents=True)
        self.lock_file_path.write_text(self._current_process_lock_file_data.to_text(), encoding='utf-8', errors='ignore')
        self._lock_succeded = True

    def on_other_instance_running(self) -> None:
        other_instance_data = self.get_existing_lock_file_data()
        raise ApplicationInstanceAlreadyRunningError(app_name=other_instance_data.app_name, running_pid=other_instance_data.pid)

    def aquire(self):
        with _SAFETY_LOCK:
            if self.lock_file_exists is True:
                existing_instance_data = self.get_existing_lock_file_data()

                if existing_instance_data.is_running is True and existing_instance_data.app_name == self._app_name:
                    self.on_other_instance_running()
                else:

                    self.store_in_lock_file()
            else:
                self.store_in_lock_file()

    def release(self):
        with _SAFETY_LOCK:
            if self._lock_succeded is True:
                self.lock_file_path.unlink(missing_ok=True)

    def __enter__(self) -> Self:
        self.aquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not ApplicationInstanceAlreadyRunningError:
            self.release()

    def __del__(self) -> None:
        self.release()


class GUISingleRunningInstanceRestrictor(SingleRunningInstanceRestrictor):

    def on_other_instance_running(self):

        other_instance_data = self.get_existing_lock_file_data()
        show_basic_error_pop_up(f"A different instance of {self._app_name!r} is already running!")

        raise ApplicationInstanceAlreadyRunningError(app_name=other_instance_data.app_name, running_pid=other_instance_data.pid)


# region [Main_Exec]
if __name__ == '__main__':
    ...

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Any, Literal, Callable, Optional
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import MissingOptionalDependencyError

with MissingOptionalDependencyError.try_import("gidapptools"):
    from rich import inspect as rinspect
    from rich import terminal_theme
    from rich.console import Console as RichConsole

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def make_dprint(**console_kwargs) -> Callable:
    use_seperator: bool = console_kwargs.pop('use_seperator', True)
    seperator_char: str = console_kwargs.pop('seperator_char', '-')
    use_print_count: bool = console_kwargs.pop('use_print_count', True)

    extra_newline: bool = console_kwargs.pop('extra_newline', True)

    prefix: Optional[str] = console_kwargs.pop('prefix', None)

    console_kwargs = {'soft_wrap': True} | console_kwargs
    console = RichConsole(**console_kwargs)
    print_count = 0

    convert_table = {'odict_keys': list, }

    def _convert_stupid_fake_iterables(in_object: Any):
        try:
            if in_object.__class__.__name__ in convert_table:
                return convert_table.get(in_object.__class__.__name__)(in_object)
        except Exception:
            return in_object
        return in_object

    def _dprint(*args, **kwargs) -> None:
        nonlocal print_count
        print_count += 1
        if use_seperator is True:
            title = "" if use_print_count is False else f"Output No.{print_count}"
            console.rule(title=title, characters=seperator_char)
        if extra_newline is True:
            kwargs['end'] = '\n\n'
        if prefix:
            console.print(prefix, end='')
        args = list(map(_convert_stupid_fake_iterables, args))
        console.print(*args, **kwargs)

    return _dprint


dprint = make_dprint()


def obj_inspection(obj: object, out_dir: Path = None, out_type: Literal["txt", "html"] = 'html', width: int = None, theme: terminal_theme.TerminalTheme = None, **rinspect_kwargs) -> None:
    console = RichConsole(soft_wrap=True, record=True)
    theme = theme or terminal_theme.MONOKAI
    rinspect_kwargs = {"methods": True, "help": True, "private": True} | rinspect_kwargs
    rinspect(obj=obj, console=console, **rinspect_kwargs)
    out_dir = Path.cwd() if out_dir is None else Path(out_dir)

    name = obj.__qualname__ or obj.__name__
    try:
        name = obj.__qualname__ or obj.__name__
    except AttributeError:
        name = obj.__name__

    out_file = out_dir.joinpath(f"{name}.{out_type}")
    if out_type == "html":
        console.save_html(out_file, theme=theme)
    elif out_type == "txt":
        console.save_text(out_file)


def is_dunder_name(in_name: str) -> bool:
    return in_name.startswith("__") and in_name.endswith("__")


    # region [Main_Exec]
if __name__ == '__main__':
    from PySide6.QtWidgets import QMenu
    obj_inspection(QMenu)

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
import pkgutil
import importlib
import os
import importlib.util
import importlib.metadata
from types import ModuleType
from pathlib import Path
from typing import Union, Iterable, Mapping


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.helper import PackageMetadataDict

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def is_importable(package_name: str) -> bool:
    """
    Checks if the package is importable, without actually trying to import it.

    :param package_name: Name of the Package to check, is case-sensitive
    :type package_name: str
    :return: True if the package is importable in the current environment
    :rtype: bool
    """
    return importlib.util.find_spec(name=package_name) is not None


def meta_data_from_module(in_module: ModuleType) -> PackageMetadataDict:
    module_metadata = importlib.metadata.metadata(in_module.__package__)
    return PackageMetadataDict.from_meta_importlib_meta_data(module_metadata)


def import_from_name(name: str) -> ModuleType:
    module = importlib.import_module(name)
    return module


def import_from_file_path(file_path: Union[str, os.PathLike, Path],
                          add_to_global_modules: bool = False) -> ModuleType:
    file_path = Path(file_path).resolve()
    _name = file_path.stem
    spec = importlib.util.spec_from_file_location(_name, file_path)
    _module = importlib.util.module_from_spec(spec)

    if add_to_global_modules is True:
        sys.modules[_name] = _module

    spec.loader.exec_module(_module)

    return _module


def all_importable_package_names(exclude_underscored: bool = True, exclude_std_lib: bool = False, exclude_main: bool = True) -> tuple[str]:

    def _check_exclude(in_name: str) -> bool:
        if exclude_underscored is True and in_name.startswith("_"):
            return False
        if exclude_std_lib is True and in_name in sys.stdlib_module_names:
            return False
        if exclude_main is True and in_name == "__main__":
            return False
        return True

    return tuple(sorted([i.name for i in pkgutil.iter_modules() if _check_exclude(i.name)], key=lambda x: x.casefold()))

# region [Main_Exec]


if __name__ == '__main__':
    for x in all_importable_package_names():
        if "css" in x.casefold():
            print(x)
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

import os
import re
import sys
import json
import queue
import math
import base64
import pickle
import random
import shelve
import dataclasses
import shutil
import asyncio
import logging
import sqlite3
import platform

import subprocess
import inspect

from time import sleep, process_time, process_time_ns, perf_counter, perf_counter_ns
from io import BytesIO, StringIO
from abc import ABC, ABCMeta, abstractmethod
from copy import copy, deepcopy
from enum import Enum, Flag, auto, unique
from pprint import pprint, pformat
from pathlib import Path
from string import Formatter, digits, printable, whitespace, punctuation, ascii_letters, ascii_lowercase, ascii_uppercase
from timeit import Timer
from typing import (TYPE_CHECKING, TypeVar, TypeGuard, TypeAlias, Final, TypedDict, Generic, Union, Optional, ForwardRef, final,
                    no_type_check, no_type_check_decorator, overload, get_type_hints, cast, Protocol, runtime_checkable, NoReturn, NewType, Literal, AnyStr, IO, BinaryIO, TextIO, Any)
from collections import Counter, ChainMap, deque, namedtuple, defaultdict
from collections.abc import (AsyncGenerator, AsyncIterable, AsyncIterator, Awaitable, ByteString, Callable, Collection, Container, Coroutine, Generator,
                             Hashable, ItemsView, Iterable, Iterator, KeysView, Mapping, MappingView, MutableMapping, MutableSequence, MutableSet, Reversible, Sequence, Set, Sized, ValuesView)
from zipfile import ZipFile, ZIP_LZMA
from datetime import datetime, timezone, timedelta
from tempfile import TemporaryDirectory
from textwrap import TextWrapper, fill, wrap, dedent, indent, shorten
from functools import wraps, partial, lru_cache, singledispatch, total_ordering, cached_property, cache
from contextlib import contextmanager, asynccontextmanager, nullcontext, closing, ExitStack, suppress
from statistics import mean, mode, stdev, median, variance, pvariance, harmonic_mean, median_grouped
from urllib.parse import urlparse
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, wait, as_completed, ALL_COMPLETED, FIRST_EXCEPTION, FIRST_COMPLETED


import PySide6
from PySide6 import (QtCore, QtGui, QtWidgets, Qt3DAnimation, Qt3DCore, Qt3DExtras, Qt3DInput, Qt3DLogic, Qt3DRender, QtAxContainer, QtBluetooth,
                     QtCharts, QtConcurrent, QtDataVisualization, QtDesigner, QtHelp, QtMultimedia, QtMultimediaWidgets, QtNetwork, QtNetworkAuth,
                     QtOpenGL, QtOpenGLWidgets, QtPositioning, QtPrintSupport, QtQml, QtQuick, QtQuickControls2, QtQuickWidgets, QtRemoteObjects,
                     QtScxml, QtSensors, QtSerialPort, QtSql, QtStateMachine, QtSvg, QtSvgWidgets, QtTest, QtUiTools, QtWebChannel, QtWebEngineCore,
                     QtWebEngineQuick, QtWebEngineWidgets, QtWebSockets, QtXml)

from PySide6.QtCore import (QByteArray, QCoreApplication, QDate, QDateTime, QEvent, QLocale, QMetaObject, QModelIndex, QModelRoleData, QMutex,
                            QMutexLocker, QObject, QPoint, QRect, QRecursiveMutex, QRunnable, QSettings, QSize, QThread, QThreadPool, QTime, QUrl,
                            QWaitCondition, Qt, QAbstractItemModel, QAbstractListModel, QAbstractTableModel, Signal, Slot)

from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient, QCursor, QFont, QFontDatabase, QFontMetrics, QGradient, QIcon, QImage,
                           QKeySequence, QLinearGradient, QPainter, QPalette, QPixmap, QRadialGradient, QTransform)

from PySide6.QtWidgets import (QApplication, QBoxLayout, QCheckBox, QColorDialog, QColumnView, QComboBox, QDateTimeEdit, QDialogButtonBox,
                               QDockWidget, QDoubleSpinBox, QFontComboBox, QFormLayout, QFrame, QGridLayout, QGroupBox, QHBoxLayout, QHeaderView,
                               QLCDNumber, QLabel, QLayout, QLineEdit, QListView, QListWidget, QMainWindow, QMenu, QMenuBar, QMessageBox,
                               QProgressBar, QProgressDialog, QPushButton, QSizePolicy, QSpacerItem, QSpinBox, QStackedLayout, QStackedWidget,
                               QStatusBar, QStyledItemDelegate, QSystemTrayIcon, QTabWidget, QTableView, QTextEdit, QTimeEdit, QToolBox, QTreeView,
                               QVBoxLayout, QWidget, QAbstractItemDelegate, QAbstractItemView, QAbstractScrollArea, QRadioButton, QFileDialog, QButtonGroup)


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    ...

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class SeparatorLine(QFrame):

    def __init__(self, orientation: Qt.Orientation = Qt.Orientation.Horizontal, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setObjectName(self.__class__.__name__)
        self._orientation: Qt.Orientation = orientation
        self.setFrameShadow(QFrame.Sunken)
        self._update_line_orientation()

    @property
    def orientation(self) -> Qt.Orientation:
        return self._orientation

    @orientation.setter
    def orientation(self, value: Qt.Orientation) -> None:
        self.set_orientation(value)

    def _update_line_orientation(self) -> None:
        shape = QFrame.Shape.HLine if self._orientation is Qt.Orientation.Horizontal else QFrame.Shape.VLine
        self.setFrameShape(shape)
        self.update()

    def set_orientation(self, orientation: Qt.Orientation):
        self._orientation = orientation
        self._update_line_orientation()

    def __repr__(self) -> str:
        return f"{__class__.__name__}(orientation={self.orientation!r}, parent={self.parent()})"


class HorizontalSeparatorLine(SeparatorLine):

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(Qt.Orientation.Horizontal, parent)


class VerticalSeparatorLine(SeparatorLine):

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(Qt.Orientation.Vertical, parent)

# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
import inspect
from typing import Any, Union, Mapping, Callable, Hashable, Iterable, Optional, Generator
from pathlib import Path
from threading import Lock
from collections import UserDict, namedtuple

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import KeyPathError, NotMappingError, AdvancedDictError, DictMergeConflict
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.general_helper.checker import is_hashable

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def replace_dict_keys(in_dict: dict, *replacement_pairs: tuple[Hashable, Hashable]) -> dict:
    # replacement_table = {item[0]: item[1] for item in replacement_pairs}
    # return in_dict.__class__(**{replacement_table.get(k, k): v for k, v in in_dict.items()})

    for old_key, new_key in replacement_pairs:
        if old_key in in_dict:
            in_dict[new_key] = in_dict.pop(old_key)

    return in_dict


def get_by_keypath(the_dict: dict, key_path: list[str], default: Any = None, *, strict: bool = False) -> Any:
    key_path = key_path.copy()
    last_key = key_path.pop(-1)
    for key in key_path:
        try:
            the_dict = the_dict[key]
        except KeyError as e:
            if strict is True:
                raise KeyError(f"The {key_path.index(key)+1}. key {key!r} was not found in the dict.") from e
            return default
    return the_dict.get(last_key, default)


def set_by_key_path(the_dict: dict, key_path: list[str], value: Any, *, create_intermediates: bool = False) -> None:
    key_path = key_path.copy()
    last_key = key_path.pop(-1)
    for key in key_path:
        try:
            the_dict = the_dict[key]
        except KeyError as e:
            if create_intermediates is True:
                the_dict[key] = {}
                the_dict = the_dict[key]
            else:
                raise KeyError(f"The {key_path.index(key)+1}. key {key!r} was not found in the dict, use 'create_intermediates=True' to auto-create intermediated keys if missing.") from e

    the_dict[last_key] = value


def multiple_dict_get(key: Hashable, *dictionaries, final_default: Any = None) -> Any:
    for _dict in dictionaries:
        value = _dict.get(key, MiscEnum.NOT_FOUND)
        if value is not MiscEnum.NOT_FOUND:
            return value
    return final_default


def multiple_dict_pop(key: Hashable, *dictionaries, final_default: Any = None) -> Any:
    for _dict in dictionaries:
        value = _dict.pop(key, MiscEnum.NOT_FOUND)
        if value is not MiscEnum.NOT_FOUND:
            return value
    return final_default


RAW_KEYPATH_TYPE = Union[list[str], str, Hashable]
MODDIFIED_KEYPATH_TYPE = list[Hashable]

ResolveKeyPathResult = namedtuple("ResolveKeyPathResult", ["data", "last_key"], defaults=(None,))


class AdvancedDict(UserDict):
    empty_values = [[], "", set(), tuple(), frozenset(), dict(), b"", None]

    def __init__(self,
                 data: Mapping = None,
                 keypath_separator: str = '.',
                 case_insensitive: bool = False,
                 convert_keys_to_str: bool = False,
                 auto_set_missing: bool = False,
                 empty_is_missing: bool = False,
                 extra_empty_values: Iterable[Any] = None) -> None:
        self.keypath_separator = keypath_separator
        self._case_insensitive = case_insensitive
        self.auto_set_missing = auto_set_missing
        self.convert_keys_to_str = convert_keys_to_str
        self.empty_is_missing = empty_is_missing
        if extra_empty_values is not None:
            self.empty_values.extend(extra_empty_values)
        self._data = data

    @property
    def data(self) -> dict:
        return self._data

    @property
    def case_insensitive(self) -> bool:
        return self._case_insensitive

    def _modify_key(self, key: Hashable) -> Hashable:

        if self.convert_keys_to_str is True and not isinstance(key, str):
            key = str(key)
        if self.case_insensitive is True and isinstance(key, str):
            return key.casefold()
        if not is_hashable(key):
            # TODO: better custom error
            raise TypeError(f"unhashable key {key!r} in key_path.")
        return key

    def _handle_keypath(self, key_path: RAW_KEYPATH_TYPE) -> MODDIFIED_KEYPATH_TYPE:
        if not isinstance(key_path, (str, Iterable)):
            return [self._modify_key(key_path)]

        if isinstance(key_path, str):
            key_path = key_path.split(self.keypath_separator)

        new_key_path = [self._modify_key(key) for key in key_path]

        return new_key_path

    def _resolve_key_path(self, key_path: MODDIFIED_KEYPATH_TYPE, *, hold_last: bool = False, check_auto_set: bool = False) -> ResolveKeyPathResult:
        data = self.data
        key_path = self._handle_keypath(key_path)
        last_key = None if hold_last is False else key_path.pop(-1)
        for key in key_path:
            try:
                if not isinstance(data, Mapping):
                    raise NotMappingError(key, data, key_path, 'get', last_key)
                data = data[key]
            except KeyError as error:
                if check_auto_set is True and self.auto_set_missing is True:
                    data[key] = {}
                    data = data[key]
                else:
                    raise KeyPathError(key, key_path, last_key) from error

        return ResolveKeyPathResult(data, last_key)

    def __getitem__(self, key_path: RAW_KEYPATH_TYPE) -> Any:
        result = self._resolve_key_path(key_path).data
        if self.empty_is_missing is True and result in self.empty_values:
            raise KeyPathError(key_path[-1], key_path)
        return result

    def __setitem__(self, key_path: RAW_KEYPATH_TYPE, value: Any) -> None:
        data, last_key = self._resolve_key_path(key_path, hold_last=True, check_auto_set=True)
        if not isinstance(data, Mapping):
            raise NotMappingError(last_key, data, self._handle_keypath(key_path), 'set')
        data[last_key] = value

    def set(self, key_path: RAW_KEYPATH_TYPE, value: Any) -> None:
        self.__setitem__(key_path=key_path, value=value)

    def __delitem__(self, key_path: RAW_KEYPATH_TYPE) -> None:
        data, last_key = self._resolve_key_path(key_path, hold_last=True)
        del data[last_key]

    def update(self, data: Mapping = None, **kwargs) -> None:
        data = {} if data is None else data
        update_data = {self._modify_key(key): value for key, value in data.items()}
        update_kwargs = {self._modify_key(key): value for key, value in kwargs.items()}
        super().update(update_data, **update_kwargs)

    # pylint: disable=arguments-renamed
    def get(self, key_path: Any, default: Any = None) -> Any:
        try:
            return self[key_path]
        except (KeyError, AdvancedDictError):
            return default

    def walk(self, temp_copy: bool = False) -> Generator[tuple[tuple[str], Any], None, None]:

        def _walk(data: dict[Hashable, Any], path: list[str] = None) -> Generator[tuple[tuple[str], Any], None, None]:
            path = [] if path is None else path
            for key, value in data.items():

                temp_path = path + [key]
                if isinstance(value, (self.__class__, dict, Mapping)):
                    yield from _walk(value, temp_path)
                else:

                    yield temp_path, value

        _data = self.data if temp_copy is False else self.data.copy()
        yield from _walk(_data)

    def modify_with_visitor(self, visitor: "BaseVisitor") -> None:
        for key_path, value in self.walk(temp_copy=True):
            visitor.visit(self, key_path, value)


class BaseVisitor:
    handle_prefix = "_handle_"
    handler_regex = re.compile(rf"^{handle_prefix}(?P<target>\w+)$")
    named_args_doc_identifier = "NAMED_VALUE_ARGUMENTS"
    _extra_global_handlers: dict[Hashable, Callable] = {}

    def __init__(self, extra_handlers: dict[Hashable, Callable] = None, default_handler: Callable = None) -> None:
        self._extra_handlers: dict[Hashable, Callable] = extra_handlers or {}
        self.default_handler = default_handler
        self._handlers: dict[Hashable, Callable] = None
        self._inspect_lock = Lock()

    @property
    def extra_handlers(self) -> dict[Hashable, Callable]:
        return self._extra_global_handlers | self._extra_handlers

    def reload(self) -> None:
        self._collect_handlers()

    @classmethod
    def _validate_new_handler(cls, handle_func: Callable) -> bool:
        # TODO: add validation logic.
        return True

    @classmethod
    def add_global_handler(cls, target_name: str, handler: Callable) -> None:
        # TODO: make it possible to inject self, or decide if.
        if cls._validate_new_handler(handler) is True:
            cls._extra_global_handlers[target_name] = handler

    def add_handler(self, target_name: str, handler: Callable) -> None:
        if self._validate_new_handler(handler) is True:
            self._extra_handlers[target_name] = handler

    def set_default_handler(self, handler: Callable):
        self.add_handler(MiscEnum.DEFAULT, handler)

    def get_all_handler_names(self) -> tuple[str]:
        return tuple(self.handlers)

    def get_all_handlers_with_named_arguments(self) -> dict[str, Optional[dict[str, str]]]:
        def get_named_args(text: str) -> dict[str, str]:
            _named_args = {}
            named_args_index = text.find(self.named_args_doc_identifier)
            next_line_index = text.find('\n', named_args_index)
            if named_args_index == -1:
                return _named_args
            gen = (line for line in text[next_line_index:].splitlines() if line)
            line = next(gen)
            while line.startswith('\t') or line.startswith(' ' * 4):
                line = line.strip()
                if not line:
                    continue
                if line in {"Returns:", "Args:"}:
                    break
                if line.strip().casefold == 'none':
                    return _named_args

                name, description = line.split(':')
                _named_args[name.strip()] = description.strip()

                line = next(gen, '')
            return _named_args

        _out = {}
        for handler_name, handler_obj in self.handlers.items():
            doc_text = handler_obj.__doc__
            if doc_text is None:
                _out[handler_name] = {}
            else:
                _out[handler_name] = get_named_args(doc_text)
        return _out

    @property
    def handlers(self) -> dict[Hashable, Callable]:
        if self._inspect_lock.locked() is True:
            return
        if self._handlers is None:
            self._collect_handlers()
        return self._handlers | self.extra_handlers

    def _collect_handlers(self) -> None:
        with self._inspect_lock:
            collected_handlers = {}
            for meth_name, meth_obj in inspect.getmembers(self, inspect.ismethod):

                match = self.handler_regex.match(meth_name)
                if match:
                    target = match.group('target')
                    if not target.strip():
                        continue
                    if target == 'default':
                        target = MiscEnum.DEFAULT
                        if self.default_handler is None:
                            self.default_handler = meth_obj

                    collected_handlers[target] = meth_obj
            if self._handlers is None:
                self._handlers = {}
            self._handlers |= collected_handlers

    def _modify_value(self, value: Any) -> Any:
        return value

    def visit(self, in_dict: Union["AdvancedDict", dict], key_path: tuple[str], value: Any) -> None:

        key_path = tuple(key_path)
        value_key = self._modify_value(value)

        handler = self.handlers.get(key_path, self.handlers.get(value_key, self.default_handler))
        if handler is None:
            return
        if isinstance(in_dict, AdvancedDict):
            in_dict.set(key_path, handler(value))
        else:
            set_by_key_path(in_dict, key_path, handler(value))


class SafeMergeDict(UserDict):
    extra_none_values: set[Any] = set()

    def __init__(self, __dict=None, raise_on_overwrite: bool = False, **kwargs) -> None:
        self.data = {} if __dict is None else __dict
        self.data |= kwargs
        self.raise_on_overwrite = raise_on_overwrite

    @classmethod
    def add_none_value(cls, none_value: Any) -> None:
        cls.extra_none_values.add(none_value)

    @classmethod
    def remove_none_value(cls, none_value: Any) -> None:
        if none_value in cls.extra_none_values:
            cls.extra_none_values.remove(none_value)

    @property
    def none_values(self) -> set[Any]:
        return {None}.union(self.extra_none_values)

    def safe_merge(self, first_dict: dict, second_dict: dict) -> dict:
        if all(key not in first_dict for key in second_dict):
            return self.__class__(first_dict | second_dict)
        new_dict = first_dict.copy()
        for key, value in second_dict.items():

            if key not in new_dict or new_dict[key] in self.none_values:
                new_dict[key] = value
            elif self.raise_on_overwrite is True:
                raise DictMergeConflict(first_dict=first_dict, second_dict=second_dict, conflicting_key=key, none_values=self.none_values)
        return self.__class__(new_dict)

    def __or__(self, other):
        if isinstance(other, UserDict):
            return self.safe_merge(self.data, other.data)
        if isinstance(other, dict):
            return self.safe_merge(self.data, other)
        return NotImplemented

    def __ror__(self, other):
        if isinstance(other, UserDict):
            return self.safe_merge(other.data, self.data)
        if isinstance(other, dict):
            return self.safe_merge(other, self.data)
        return NotImplemented

    def __ior__(self, other):
        if isinstance(other, UserDict):
            self.data = self.safe_merge(self.data, other.data)
        else:
            self.data = self.safe_merge(self.data, other)
        return self


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
from typing import TYPE_CHECKING
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QScreen
from PySide6.QtWidgets import QWidget, QApplication

if sys.version_info >= (3, 11):
    pass
else:
    pass
# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    ...

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def move_to_center_of_screen(widget: QWidget, screen: QScreen = None) -> None:
    screen = screen or QApplication.instance().screenAt(widget.cursor().pos())

    screen_center = screen.availableGeometry().center()
    widget_frame_geometry = widget.frameGeometry()

    widget_frame_geometry.moveCenter(screen_center)

    # widget.setGeometry(widget_frame_geometry)

    # widget.move(screen_center)


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING, Any, Union, Literal, Callable
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import EntryMissingError, SectionExistsError, SectionMissingError
from gidapptools.general_helper.timing import get_dummy_profile_decorator_in_globals
from gidapptools.gid_config.parser.tokens import Entry, Section
from gidapptools.general_helper.class_helper import MethodEnabledWeakSet
from gidapptools.gid_config.parser.ini_parser import BaseIniParser
from gidapptools.general_helper.mixins.file_mixin import FileMixin

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.custom_types import PATH_TYPE

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]
get_dummy_profile_decorator_in_globals()
THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class ConfigData:

    def __init__(self, name: str) -> None:
        self.name = name
        self._sections: dict[str, Section] = None

    @property
    def sections(self) -> dict[str, Section]:
        if self._sections is None:
            self._sections = {}
        return self._sections

    @property
    def all_sections(self) -> tuple[Section]:
        return tuple(self.sections.values())

    @property
    def all_section_names(self) -> tuple[str]:
        return tuple(self.sections)

    def has_section(self, section_name: str) -> bool:
        return section_name in self.all_section_names

    def has_key(self, section_name: str, key_name: str) -> bool:
        if self.has_section(section_name=section_name) is False:
            return False

        return self.get_section(section_name=section_name).has_key(key_name=key_name)

    def get_section(self, section_name: str, create_missing_section: bool = True) -> Section:
        try:
            return self.sections[section_name]
        except (KeyError, SectionMissingError) as error:
            if create_missing_section is False:
                raise SectionMissingError(section_name=section_name, config_data=self) from error
            section = Section(section_name)
            self.add_section(section=section)
            return section

    def add_section(self, section, existing_ok: bool = True) -> bool:
        if section.name in self._sections:
            if existing_ok is True:
                return False
            raise SectionExistsError(f"Section {section.name!r} already exists.")
        self._sections[section.name] = section
        return True

    def remove_section(self, section_name: str, missing_ok: bool = False) -> bool:
        try:
            del self._sections[section_name]
            return True
        except KeyError as error:
            if missing_ok is False:
                raise SectionMissingError(section_name=section_name, config_data=self) from error
            return False

    def clear_all_sections(self) -> bool:
        self._sections = None
        return True

    def get_entry(self, section_name: str, entry_key: str, create_missing_section: bool = True) -> Entry:
        section = self.get_section(section_name=section_name, create_missing_section=create_missing_section)
        try:
            return section[entry_key]
        except KeyError as error:
            raise EntryMissingError(section_name=section_name, entry_key=entry_key, config_data=self) from error

    def add_entry(self, section_name: str, entry: Entry, create_missing_section: bool = True) -> bool:
        section = self.get_section(section_name=section_name, create_missing_section=create_missing_section)
        section.add_entry(entry=entry)
        return True

    def set_value(self, section_name: str, entry_key: str, entry_value: str, create_missing_section: bool = True) -> bool:
        try:
            entry = self.get_entry(section_name=section_name, entry_key=entry_key, create_missing_section=create_missing_section)
            entry.value = entry_value

        except EntryMissingError:
            entry = Entry(entry_key, entry_value)
            self.add_entry(section_name=section_name, entry=entry, create_missing_section=create_missing_section)
        return True

    def remove_entry(self, section_name: str, entry_key: str, missing_ok: bool = False) -> bool:
        try:
            section = self.get_section(section_name=section_name)
        except SectionMissingError:
            if missing_ok is False:
                raise
            return False

        try:
            del section[entry_key]
            return True
        except KeyError as error:
            if missing_ok is False:
                raise EntryMissingError(section_name=section_name, entry_key=entry_key, config_data=self) from error
            return False

    def clear_entries(self, section_name: str, missing_ok: bool = False) -> bool:
        try:
            section = self.get_section(section_name=section_name)
            section.entries = {}
            return True
        except SectionMissingError:
            if missing_ok is False:
                raise
            return False

    def reload(self) -> None:
        pass

    def as_raw_dict(self) -> dict[str, dict[str, Any]]:
        _out = {}
        sections = self.sections.copy()
        for section in sections.values():

            _out |= section.as_dict()
        return _out


class ConfigFile(FileMixin, ConfigData):
    _name_suffixes_to_remove: tuple[str] = ("_configspec", "_spec", "_config", "_config_spec", "_spec_config", "_specconfig")

    def __init__(self,
                 file_path: "PATH_TYPE",
                 parser: BaseIniParser,
                 changed_parameter: Union[Literal['size'], Literal['file_hash']] = 'size',
                 auto_write: bool = True,
                 missing_ok: bool = True,
                 **kwargs) -> None:

        self.parser = parser
        self.auto_write = auto_write
        super().__init__(name=self._generate_name_from_path(file_path, suffixes_to_remove=self._name_suffixes_to_remove),
                         file_path=file_path,
                         changed_parameter=changed_parameter,
                         missing_ok=missing_ok,
                         **kwargs)
        self._on_reload_targets: MethodEnabledWeakSet = MethodEnabledWeakSet()

    def _do_auto_write(self, success: bool) -> None:
        if success is True and self.auto_write is True:
            self.save()

    @property
    def sections(self) -> dict[str, Section]:
        if self._sections is None or self.has_changed is True:
            self.reload()
        return self._sections

    def add_on_reload_target(self, target: Callable[[], None]) -> None:
        self._on_reload_targets.add(target)

    def reload_if_changed(self) -> None:
        with self.lock:
            if self.has_changed is True:
                self.reload()

    def reload(self) -> None:
        with self.lock:
            self.load()
            for target in self._on_reload_targets:
                target()

    def set_value(self, section_name: str, entry_key: str, entry_value: str, create_missing_section: bool = True) -> bool:
        success = super().set_value(section_name, entry_key, entry_value, create_missing_section=create_missing_section)
        self._do_auto_write(success)
        return success

    def add_entry(self, section_name: str, entry: Entry, create_missing_section: bool = True) -> bool:
        success = super().add_entry(section_name, entry, create_missing_section=create_missing_section)
        self._do_auto_write(success)
        return success

    def add_section(self, section, existing_ok: bool = True) -> None:
        self.reload_if_changed()
        success = super().add_section(section, existing_ok=existing_ok)
        self._do_auto_write(success)
        return success

    def remove_entry(self, section_name: str, entry_key: str, missing_ok: bool = False) -> bool:
        success = super().remove_entry(section_name, entry_key, missing_ok=missing_ok)
        self._do_auto_write(success)
        return success

    def remove_section(self, section_name: str, missing_ok: bool = False) -> bool:
        success = super().remove_section(section_name, missing_ok=missing_ok)
        self._do_auto_write(success)
        return success

    def clear_all_sections(self) -> bool:
        success = super().clear_all_sections()
        self._do_auto_write(success)
        return success

    def clear_entries(self, section_name: str, missing_ok: bool = False) -> bool:
        success = super().clear_entries(section_name, missing_ok=missing_ok)
        self._do_auto_write(success)
        return success

    def save(self) -> None:
        with self.lock:
            sections = self.sections.copy()

            data = '\n\n'.join(section.as_text() for section in sections.values())

            self.write(data)

    def load(self) -> None:
        with self.lock:
            content = self.read()
            self._sections = {section.name: section for section in self.parser.parse(content)}
# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Union
from pathlib import Path
from concurrent.futures import Future

# * Qt Imports --------------------------------------------------------------------------------------->
import PySide6
from PySide6.QtGui import QMovie, QColor
from PySide6.QtCore import Qt, QSize, Signal
from PySide6.QtWidgets import QLabel, QWidget, QPushButton, QSizePolicy, QVBoxLayout, QApplication

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.data.gifs import StoredGif, get_gif

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class BusySpinnerWidget(QLabel):
    default_gif_name: str = "busy_spinner_4.gif"
    default_spinner_size: tuple[int, int] = (75, 75)
    _stop_signal = Signal(Future)

    def __init__(self, parent: QWidget = None, spinner_gif: Union[QMovie, str, Path, StoredGif] = None, spinner_size: QSize = None):
        super().__init__(parent)
        self.spinner_size = spinner_size or QSize(*self.default_spinner_size)
        self.spinner_gif_item, self.spinner_gif = self.setup_spinner_gif(spinner_gif=spinner_gif)
        self.setAlignment(Qt.AlignCenter)
        self._stop_signal.connect(self.stop)
        self.running: bool = False

    @property
    def app(self) -> QApplication:
        return QApplication.instance()

    def set_spinner_size(self, size: QSize) -> None:
        self.spinner_size = size
        self.spinner_gif.setScaledSize(self.spinner_size)

    def setup_spinner_gif(self, spinner_gif: Union[QMovie, str, Path, StoredGif]) -> tuple[StoredGif, QMovie]:
        if isinstance(spinner_gif, str):
            if Path(spinner_gif).is_file() is True:
                spinner_gif = StoredGif(spinner_gif)
            else:
                spinner_gif = get_gif(spinner_gif)
        elif isinstance(spinner_gif, Path):
            spinner_gif = StoredGif(spinner_gif)

        spinner_gif_item = spinner_gif or get_gif(self.default_gif_name)
        spinner_gif = QMovie(str(spinner_gif_item.path))

        spinner_gif.setScaledSize(self.spinner_size)
        spinner_gif.setCacheMode(QMovie.CacheAll)
        self.setMovie(spinner_gif)
        return spinner_gif_item, spinner_gif

    def start(self):
        self.spinner_gif.start()
        self.running = True

    def stop(self):
        self.spinner_gif.stop()
        self.running = False


class BusyPushButton(QPushButton):

    def __init__(self,
                 parent: QWidget = None,
                 text: str = None,
                 spinner_gif: Union[QMovie, str] = None,
                 spinner_size: QSize = None,
                 disable_while_spinning: bool = True):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Minimum)
        self.setLayout(QVBoxLayout())
        self.disable_while_spinning = disable_while_spinning
        self.text_widget = QLabel(self)
        self.text_widget.setScaledContents(True)
        self.text_widget.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(self.text_widget)
        self.busy_spinner_widget = BusySpinnerWidget(self, spinner_gif=spinner_gif, spinner_size=spinner_size or QSize(self.sizeHint().height(), self.sizeHint().height()))
        self.layout.addWidget(self.busy_spinner_widget)
        self.set_text(text)
        self.busy_spinner_widget.setVisible(False)
        self.busy_spinner_widget._stop_signal.connect(self.stop_spinner)

    @property
    def layout(self) -> QVBoxLayout:
        return super().layout()

    def resize(self, size: QSize):
        super().resize(size)
        self.set_spinner_size(QSize(self.size().height(), self.size().height()))

    def set_spinner_size(self, size: QSize):
        self.busy_spinner_widget.set_spinner_size(size)

    def set_text(self, text: str):
        text = text or ""
        self.text_widget.setText(text)

    def sizeHint(self) -> PySide6.QtCore.QSize:
        return self.layout.sizeHint()

    def hide_text(self):

        self.text_widget.setVisible(False)

    def show_text(self):

        self.text_widget.setVisible(True)

    def start_spinner(self):
        self.hide_text()
        self.busy_spinner_widget.setVisible(True)

        self.busy_spinner_widget.start()
        if self.disable_while_spinning is True:
            self.setEnabled(False)

    def stop_spinner(self, *args):
        if self.disable_while_spinning is True:
            self.setEnabled(True)
        self.busy_spinner_widget.stop()
        self.busy_spinner_widget.setVisible(False)
        self.show_text()

    def start_spinner_while_future(self, future: Future):
        self.start_spinner()

        future.add_done_callback(self.busy_spinner_widget._stop_signal.emit)

    def start_spinner_with_stop_signal(self, stop_signal: Signal):
        self.start_spinner()
        stop_signal.connect(self.stop_spinner)


# region [Main_Exec]
if __name__ == '__main__':
    app = QApplication()
    x = BusySpinnerWidget()
    x.show()
    app.exec()

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import ast
import inspect
import pkgutil
from types import ModuleType
from typing import Any, Optional
from pathlib import Path
from importlib import import_module
from collections.abc import Iterable
from importlib.metadata import metadata

# * Third Party Imports --------------------------------------------------------------------------------->
import attrs

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import MissingOptionalDependencyError
from gidapptools.general_helper.development.misc import is_dunder_name

with MissingOptionalDependencyError.try_import("gidapptools"):
    import isort
try:
    from rich.console import Console as RichConsole
    RICH_IMPORTABLE = True
except ImportError:
    RICH_IMPORTABLE = False

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class ClassNameFinder(ast.NodeVisitor):

    def __init__(self) -> None:
        super().__init__()
        self.class_names: list[str] = []

    def visit_ClassDef(self, node: ast.ClassDef) -> Any:
        self.class_names.append(node.name)
        self.generic_visit(node)

    def get_unique_class_names(self) -> tuple[str]:
        class_names = set(self.class_names)
        return tuple(sorted(class_names, key=lambda x: (not x.startswith("Q"), x.casefold())))


@attrs.define(frozen=True, slots=True)
class SubModule:
    name: str
    qualname: str
    package: str
    module_info: pkgutil.ModuleInfo = attrs.field(default=None)

    @classmethod
    def from_module_info(cls, in_module_info: pkgutil.ModuleInfo) -> "SubModule":
        name = in_module_info.name.split('.')[-1]
        qualname = in_module_info.name
        package = in_module_info.name.split('.')[0]

        return cls(name=name, qualname=qualname, module_info=in_module_info, package=package)

    @property
    def full_name(self) -> str:
        return self.module.__name__

    @property
    def path(self) -> Optional[Path]:
        try:
            path = self.module.__file__
            if path is not None:
                return Path(path)
        except AttributeError:
            return None

    @property
    def is_top_module(self) -> bool:
        return False

    @property
    def dunder_names(self) -> tuple[str]:
        exclude_names = {"__builtins__"}
        return tuple(name for name in dir(self.module) if is_dunder_name(name) is True and name not in exclude_names)

    @property
    def doc(self) -> str:
        doc = self.module.__doc__
        if doc is None and self.path.suffix == ".pyd":
            typing_file = self.path.with_suffix(".pyi")
            if typing_file.is_file() is True:
                tree = ast.parse(typing_file.read_text(encoding='utf-8', errors='ignore'))
                doc = ast.get_docstring(tree)

        if doc is not None:
            doc = inspect.cleandoc(doc)
        return doc

    @property
    def meta_data(self) -> dict[str, Any]:
        return dict(metadata(self.package))

    @property
    def module(self) -> ModuleType:
        return import_module(self.qualname, self.package)

    @property
    def import_string(self) -> str:
        import_path = self.qualname.rsplit('.', 1)[0]
        return f"from {import_path} import {self.name}"

    @property
    def all_members_import_string(self) -> str:
        text = f"from {self.qualname} import ({', '.join(self.member_names)})"
        try:
            return isort.code(text, line_length=200).strip()
        except Exception as e:
            print(f"Encountered exception {e!r}")

            return text.strip()

    @property
    def member_names(self) -> tuple[str]:

        def _check(_name: str, _obj: object) -> bool:
            return not any([
                getattr(_obj, "__module__", None) != self.module.__name__,
                inspect.ismodule(_obj),
                inspect.isbuiltin(_obj),
                _name.startswith("_"),
                _name.endswith("__")
            ])

        def _sort_key(in_name: str) -> tuple:
            norm_name = in_name.casefold()
            has_upper_q_prefix = in_name.startswith("Q")
            has_lower_q_prefix = in_name.startswith("q")

            return (not has_upper_q_prefix, not has_lower_q_prefix, norm_name, len(norm_name))

        _out = set()
        for name, obj in inspect.getmembers(self.module):
            if _check(name, obj) is False:
                continue

            _out.add(name)

        return tuple(sorted(_out, key=_sort_key))

    def to_dict(self, for_json: bool = False) -> dict[str, Any]:

        def _std_serialize(inst: "SubModule", field_name: str, value: Any) -> Any:
            match value:
                case pkgutil.ModuleInfo:
                    raise RuntimeError("Module info should not hit this")
            return value

        def _json_serialize(inst: "SubModule", field_name: str, value: Any) -> Any:
            match value:

                case Path():
                    return value.as_posix()

                case float():
                    return str(value)

                case tuple():
                    return [_json_serialize(inst, field_name, i) for i in value]

                case set():
                    return [_json_serialize(inst, field_name, i) for i in value]

                case str():
                    return value

                case list():
                    return [_json_serialize(inst, field_name, i) for i in value]

                case dict():
                    return {_json_serialize(inst, field_name, k): _json_serialize(inst, field_name, v) for k, v in value.items()}

                case int():
                    return value

                case None:
                    return value

                case _:
                    return str(value)

        def _filter(field_name: str, value: Any) -> bool:
            match field_name:
                case "module_info":
                    return False

                case "module":
                    return False

            return True

        def _result_sorting_key(name_value: tuple[str, object]) -> Optional[int]:
            fixed_order = ("name",
                           "qualname",
                           "package",
                           "doc")
            try:
                pos = fixed_order.index(name_value[0])
            except ValueError:
                if isinstance(name_value[1], str):
                    pos = 1_000
                elif isinstance(name_value[1], Path):
                    pos = 2_000
                elif isinstance(name_value[1], Iterable):
                    pos = 3_000
                else:
                    pos = 10_000
            return pos

        _out = {}

        attrs_attribute_names = [field.name for field in attrs.fields(self.__class__)]
        extra_attribute_names = ["path", "doc", "module", "import_string", "all_members_import_string", "member_names", "dunder_names"]

        attribute_names = attrs_attribute_names + extra_attribute_names

        _serialize = _std_serialize if for_json is False else _json_serialize

        for name in attribute_names:
            value = getattr(self, name)
            if _filter(name, value) is False:
                continue
            serialized_value = _serialize(self, name, value)
            _out[name] = serialized_value
        return {k: v for k, v in sorted(_out.items(), key=_result_sorting_key)}


class TopModule(SubModule):

    @property
    def is_top_module(self) -> bool:
        return True


class PackageInfoData:

    def __init__(self, module: ModuleType) -> None:
        self._sub_modules_data: dict[str, "SubModule"] = dict()
        self._top_module_data: "TopModule" = TopModule.from_module_info(pkgutil.ModuleInfo(pkgutil.find_loader(module.__name__), module.__name__, getattr(module, "__path__", None) is not None))

    @property
    def top_module_data(self) -> TopModule:
        return self._top_module_data

    @property
    def name(self) -> str:
        return self.top_module_data.full_name

    def _insert_item(self, item: SubModule):
        if item.name != item.full_name:
            top_name = (self.name.rsplit(".", 1)[0] + ".") if "." in self.name else ""
            new_key = item.full_name.replace(top_name, "")
            self._sub_modules_data[new_key] = item
        else:
            self._sub_modules_data[item.name] = item

    def append(self, item: "SubModule") -> None:
        self._insert_item(item)
        self._sort_sub_modules()

    def add(self, items: Iterable["SubModule"]) -> None:
        for item in items:
            self._insert_item(item)
        self._sort_sub_modules()

    def _sort_key_func(self, in_item_tuple: tuple[str, "SubModule"]) -> tuple[bool, str, int]:
        return (in_item_tuple[0].startswith("_"), in_item_tuple[1].qualname, len(in_item_tuple[1].qualname.split(".")))

    def _sort_sub_modules(self) -> None:
        self._sub_modules_data = dict(sorted(self._sub_modules_data.items(), key=self._sort_key_func))

    def __iter__(self):
        return iter([self.top_module_data] + list(self._sub_modules_data.values()))

    def __getitem__(self, name: str) -> "SubModule":
        if name == self.top_module_data.name:
            return self.top_module_data

        return self._sub_modules_data[name]

    def to_dict(self, for_json: bool = False) -> list[dict[str, Any]]:
        return [i.to_dict(for_json=for_json) for i in ([self.top_module_data] + list(self._sub_modules_data.values()))]

    def __repr__(self) -> str:

        return f"{self.__class__.__name__}({self.name!r}, sub_modules={list(self._sub_modules_data.keys())!r})"


def get_all_sub_modules(in_module: ModuleType) -> PackageInfoData:
    module_data = PackageInfoData(in_module)

    try:
        module_path = in_module.__path__
        module_prefix = in_module.__package__ + "."

        for info in pkgutil.walk_packages(module_path, module_prefix):
            sub_module = SubModule.from_module_info(info)
            module_data.append(sub_module)

    except AttributeError:
        pass
    return module_data


def get_all_sub_module_data(in_module: ModuleType, for_json: bool = False) -> list[dict[str, object]]:
    return get_all_sub_modules(in_module).to_dict(for_json=for_json)


def print_all_sub_module_data(in_module: ModuleType, no_rich: bool = False) -> None:

    if no_rich is True or RICH_IMPORTABLE is False:
        from pprint import pprint
        pprint(get_all_sub_module_data(in_module=in_module))
        return

    console = RichConsole(soft_wrap=True)
    console.print_json(data=get_all_sub_module_data(in_module), default=str)


# region [Main_Exec]
if __name__ == '__main__':
    from PySide6 import QtWidgets
    import importlib.resources
    print_all_sub_module_data(importlib.resources)

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING, Optional
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.gid_config.conversion.converter_grammar import ConverterSpecData, reverse_replace_value_words

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    pass

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class SpecSection:
    __slots__ = ("name",
                 "default_converter",
                 "description",
                 "dynamic_entries_allowed",
                 "gui_visible",
                 "implemented",
                 "verbose_name",
                 "entries")

    def __init__(self,
                 name: str,
                 default_converter: ConverterSpecData = MiscEnum.NOTHING,
                 description: str = "",
                 dynamic_entries_allowed: bool = False,
                 gui_visible: bool = True,
                 implemented: bool = True,
                 verbose_name: str = MiscEnum.NOTHING) -> None:
        self.name = name
        self.default_converter = default_converter
        self.description = description
        self.dynamic_entries_allowed = dynamic_entries_allowed
        self.gui_visible = gui_visible
        self.implemented = implemented
        self.verbose_name = verbose_name if verbose_name is not MiscEnum.NOTHING else self.name.replace("_", " ").title()
        self.entries: dict[str, "SpecEntry"] = {}

    def __getitem__(self, name: str) -> "SpecEntry":
        return self.entries[name]

    def add_entry(self, entry: "SpecEntry") -> None:
        entry.set_section(self)
        self.entries[entry.name] = entry

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name={self.name!r})'


def converter_data_to_string(converter_data: "ConverterSpecData") -> str:
    text = converter_data["typus"]
    if converter_data["kw_arguments"]:
        sub_args = []
        for k, v in converter_data["kw_arguments"].items():
            sub_args.append(f"{k}={reverse_replace_value_words(v)}")

        text += "(" + ', '.join(sub_args) + ")"
    return text


class SpecEntry:
    __slots__ = ("name",
                 "converter",
                 "default",
                 "description",
                 "gui_visible",
                 "implemented",
                 "verbose_name",
                 "_section",
                 "initial_value")

    def __init__(self,
                 name: str,
                 converter: ConverterSpecData = MiscEnum.NOTHING,
                 default: str = MiscEnum.NOTHING,
                 description: str = "",
                 verbose_name: str = MiscEnum.NOTHING,
                 implemented: bool = True,
                 gui_visible: bool = True,
                 initial_value: str = MiscEnum.NOTHING) -> None:
        self.name = name
        self.converter = converter
        self.default = default
        self.description = description
        self.verbose_name = verbose_name if verbose_name is not MiscEnum.NOTHING else self.name.replace("_", " ").title()
        self.implemented = implemented
        self.gui_visible = gui_visible
        self._section: SpecSection = None
        self.initial_value = initial_value

    @property
    def section(self) -> Optional[SpecSection]:
        return self._section

    def set_section(self, section: SpecSection) -> None:
        self._section = section
        if self.converter is MiscEnum.NOTHING and self.section.default_converter is not MiscEnum.NOTHING:
            self.converter = self.section.default_converter

        if self.section.gui_visible is False:
            self.gui_visible = False

        if self.section.implemented is False:
            self.implemented = False

    def __getitem__(self, name: str):
        if name in {"name", "converter", "default", "description", "verbose_name", "implemented", "gui_visible"}:
            return getattr(self, name)
        raise KeyError(name)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name={self.name!r})'


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
from typing import TYPE_CHECKING, Union, TypeVar
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
from frozendict import frozendict
from gidapptools.gid_logger.logger import get_logger

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gid_database.orm_peewee.sqlite.apsw_database import GidAPSWDatabase

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()
log = get_logger(__name__)

# endregion [Constants]

x = {'analysis_limit': 100000,
     'application_id': 0,
     'auto_vacuum': 2,
     'automatic_index': 1,
     'busy_timeout': 0,
     'cache_size': -128000,
     'cache_spill': 30971,
     'cell_size_check': 0,
     'checkpoint_fullfsync': 0,
     'collation_list': 0,
     'data_version': 2,
     'database_list': 0,
     'defer_foreign_keys': 0,
     'encoding': 'UTF-8',
     'foreign_keys': 1,
     'freelist_count': 0,
     'fullfsync': 0,
     'hard_heap_limit': 0,
     'ignore_check_constraints': 0,
     'journal_mode': 'wal',
     'journal_size_limit': 1073741824,
     'legacy_alter_table': 0,
     'locking_mode': 'normal',
     'max_page_count': 1073741823,
     'mmap_size': 0,
     'page_count': 39215,
     'page_size': 4096,
     'query_only': 0,
     'read_uncommitted': 0,
     'recursive_triggers': 0,
     'reverse_unordered_selects': 0,
     'secure_delete': 0,
     'soft_heap_limit': 0,
     'synchronous': 0,
     'temp_store': 0,
     'threads': 8,
     'trusted_schema': 1,
     'user_version': 5,
     'wal_autocheckpoint': 0}


_ONLY_EXECUTABLE_PRAGMA_NAMES: tuple[str] = ("foreign_key_check",
                                             "optimize",
                                             "integrity_check",
                                             "quick_check",
                                             "wal_checkpoint",
                                             "incremental_vacuum",
                                             "shrink_memory")

_DEPRECATED_PRAGMA_NAMES: tuple[str] = ("count_changes",
                                        "data_store_directory",
                                        "default_cache_size",
                                        "empty_result_callbacks",
                                        "full_column_names",
                                        "short_column_names",
                                        "temp_store_directory",
                                        "pragma_list")

_TESTING_ONLY_PRAGMA_NAMES: tuple[str] = ("stats",
                                          "writable_schema")

_NON_QUERYABLE_PRAGMA_NAMES: tuple[str] = ("schema_version",
                                           "case_sensitive_like",
                                           "foreign_key_list",
                                           "index_info",
                                           "index_list",
                                           "index_xinfo",
                                           "table_info",
                                           "table_xinfo")

_MULTIPLE_VALUE_PRAGMAS: tuple[str] = ("table_list", "function_list", "module_list", "compile_options", "database_list")

_GET_DEFAULT_TYPE = TypeVar("_GET_DEFAULT_TYPE")


class PragmaInfo:
    _pragma_names_to_exclude: set[str] = set(_ONLY_EXECUTABLE_PRAGMA_NAMES + _DEPRECATED_PRAGMA_NAMES + _TESTING_ONLY_PRAGMA_NAMES + _NON_QUERYABLE_PRAGMA_NAMES + _MULTIPLE_VALUE_PRAGMAS)

    def __init__(self, database: "GidAPSWDatabase") -> None:
        self.database = database
        self._all_pragma_names: tuple[str] = None
        self._pragma_data: frozendict[str, Union[str, int]] = None
        self._compile_options: frozendict[str, object] = None
        self._module_list: tuple[str] = None

    @property
    def all_pragma_names(self) -> tuple[str]:
        if self._all_pragma_names is None:
            self._all_pragma_names = self._get_all_pragma_names()
        return self._all_pragma_names

    @property
    def pragma_data(self) -> frozendict[str, Union[str, int]]:
        if self._pragma_data is None:
            self._pragma_data = self._get_pragma_data()
        return self._pragma_data

    @property
    def compile_options(self) -> frozendict[str, object]:
        if self._compile_options is None:
            self._compile_options = self._get_compile_options()
        return self._compile_options

    @property
    def module_list(self) -> tuple[str]:
        if self._module_list is None:
            self._module_list = self._get_module_list()
        return self._module_list

    def _get_all_pragma_names(self) -> tuple[str]:
        return tuple(i[0] for i in self.database.connection().execute("PRAGMA pragma_list"))

    def _get_pragma_data(self) -> frozendict[str, Union[str, int]]:
        _out = {}
        pragma_names_to_query = (name for name in self.all_pragma_names if name not in self._pragma_names_to_exclude)

        for pragma_name in pragma_names_to_query:
            try:
                _out[pragma_name] = self.database.connection().execute(f"PRAGMA {pragma_name}").fetchone()[0]
            except TypeError:
                _out[pragma_name] = None
        return frozendict(_out)

    def _get_compile_options(self) -> frozendict[str, object]:
        _out = {}
        for option_entry in [i[0] for i in self.database.connection().execute("PRAGMA compile_options")]:

            if "=" in option_entry:
                name, value = (i.strip() for i in option_entry.split("=", 1))
            else:
                name = option_entry.strip()
                value = "NO_VALUE"

            _out[name] = value
        return frozendict(_out)

    def _get_module_list(self) -> tuple[str]:
        return tuple(i[0].strip() for i in self.database.connection().execute("PRAGMA module_list"))

    def fill_with_data(self) -> Self:
        self._all_pragma_names = self._get_all_pragma_names()
        self._pragma_data = self._get_pragma_data()
        self._compile_options = self._get_compile_options()
        self._module_list = self._get_module_list()
        log.info("finished filling %r with data", self)
        return self

    def reset(self) -> Self:
        self._all_pragma_names = None
        self._pragma_data = None
        self._compile_options = None
        self._module_list = None
        log.info("reset %r", self)
        return self

    def __getitem__(self, key: str) -> Union[str, int]:
        return self.pragma_data[key]

    def get(self, key: str, default: _GET_DEFAULT_TYPE = None) -> Union[_GET_DEFAULT_TYPE, int, str]:
        try:
            self[key]
        except KeyError:
            return default

    def on_pragma_set(self, pragma_name: str, value: Union[str, int, None]) -> None:
        if pragma_name in self.pragma_data:
            self._pragma_data[pragma_name] = value

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(database={self.database!r})'


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QPalette, QMouseEvent, QDesktopServices
from PySide6.QtCore import Qt, QUrl
from PySide6.QtWidgets import QLabel, QApplication, QMainWindow, QWidget, QGridLayout, QHBoxLayout, QVBoxLayout
from validators import url as validate_url
import requests
from time import perf_counter
# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class HyperlinkLabel(QLabel):

    def __init__(self, link: str = None, validate: bool = False, parent=None):
        super().__init__(parent=parent)
        self.validate = validate
        self._link: str = None
        if link:
            self.set_link(link)
        self._set_link_color()

    @property
    def link(self) -> str:
        return self._link

    def get_link(self) -> str:
        return self._link

    def get_link_as_QUrl(self) -> QUrl:
        return QUrl(self._link)

    def _validate_link(self, link: str):
        result = validate_url(link)
        if not result:
            raise result

    def _modify_link(self, link: str) -> str:
        return link

    def set_link(self, link: str):
        link = self._modify_link(link)
        if self.validate is True:
            self._validate_link(link)
        self._link = link
        self.setText(link)
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)

    def _set_link_color(self):
        link_color = QApplication.instance().palette().color(QPalette.Button.Link)
        r = link_color.red()
        g = link_color.green()
        b = link_color.blue()
        self.setStyleSheet(f"color: rgb({', '.join(str(i) for i in [r,g,b])})")
        self.setCursor(Qt.PointingHandCursor)

    def _open_link(self):
        QDesktopServices.openUrl(self._link)

    def mousePressEvent(self, ev: QMouseEvent) -> None:
        if ev.button() == Qt.LeftButton:
            self._open_link()
        else:
            super().mousePressEvent(ev)
# region [Main_Exec]


if __name__ == '__main__':
    x = QApplication()
    w = QMainWindow()
    cont_wid = HyperlinkLabel("https://example.com/", validate=True)

    cent_wid = QWidget()
    cent_wid.setLayout(QVBoxLayout())
    lab = QLabel("blah")

    cent_wid.layout().addWidget(cont_wid)

    w.setCentralWidget(cent_wid)
    url = cont_wid.get_link_as_QUrl()

    w.show()
    x.exec()

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class BaseToken:

    def __repr__(self) -> str:
        var_parts = [f"{k}={v!r}" for k, v in vars(self).items()]
        return f"{self.__class__.__name__}({', '.join(var_parts)})"


class BaseTokenWithPos(BaseToken):

    def __init__(self,
                 start: int,
                 end: int) -> None:
        self.start = start
        self.end = end
        self.span = end - start

    @classmethod
    def from_parse_action(cls, s, l, t) -> "BaseTokenWithPos":
        data_dict = t[0].as_dict()
        data_dict["start"] = data_dict.pop("locn_start")
        data_dict["end"] = data_dict.pop("locn_end")
        return cls(**data_dict)


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from enum import unique
from typing import TYPE_CHECKING, Union, AnyStr, Literal, Iterable, Callable
from hashlib import md5, sha256, blake2b, blake2s, sha3_512
from pathlib import Path
from threading import RLock

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import BaseGidEnum
from gidapptools.gid_signal.interface import get_signal
from gidapptools.general_helper.timing import get_dummy_profile_decorator_in_globals
from gidapptools.vendored.atomic_writes import atomic_write
from gidapptools.general_helper.conversion import human2bytes
from gidapptools.general_helper.concurrency.locks import GLOBAL_RLOCK_MANAGER

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.custom_types import PATH_TYPE

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]
get_dummy_profile_decorator_in_globals()
THIS_FILE_DIR = Path(__file__).parent.absolute()


# endregion [Constants]


READ_TYPE = Union[Literal["r"], Literal["rb"]]
WRITE_TYPE = Union[Literal["w"], Literal["wb"], Literal['a'], Literal['ab']]
ON_ERRORS_TYPE = Union[Literal['replace'], Literal['ignore'], AnyStr]
HASH_FUNC_TYPE = Union[blake2b, md5, sha256, sha3_512, blake2s]


class FileMixin(os.PathLike):
    _encoding = 'utf-8'
    _on_errors: ON_ERRORS_TYPE = 'ignore'

    hash_func: HASH_FUNC_TYPE = md5
    file_hash_size_threshold: int = human2bytes("100 mb")

    @ unique
    class ChangeParameter(BaseGidEnum):
        SIZE = "size"
        FILE_HASH = "file_hash"
        MTIME = "mtime"
        ALWAYS = "always"
        NEVER = "never"
        ALL = "all"

        def get_check_method(self, instance: "FileMixin") -> Callable[[], bool]:
            return getattr(instance, f"_on_{self.value}_changed_check")

    def __init__(self, file_path: "PATH_TYPE", changed_parameter: str = None, missing_ok: bool = True, **kwargs) -> None:
        self.file_path = Path(file_path)
        self.changed_parameter = self.ChangeParameter.SIZE if changed_parameter is None else self.ChangeParameter(changed_parameter)
        self.changed_check_func: Callable[[], bool] = self.changed_parameter.get_check_method(self)
        self.missing_ok = missing_ok
        self.read_mode: READ_TYPE = 'r'
        self.write_mode: WRITE_TYPE = 'w'
        self.last_size: int = None
        self.last_file_hash: str = None
        self.last_mtime: int = None
        self.changed_signal = get_signal(key=self.file_path)
        self.lock: RLock = GLOBAL_RLOCK_MANAGER.get_file_lock(self.file_path)
        self.file_was_created: bool = False
        self._check_handle_not_existing()
        super().__init__(**kwargs)

    @staticmethod
    def _generate_name_from_path(in_file_path: "PATH_TYPE", suffixes_to_remove: Iterable[str] = None) -> str:
        possible_suffixes_to_remove = tuple(suffixes_to_remove) if suffixes_to_remove is not None else tuple()
        in_file_path = Path(in_file_path)
        raw_name = in_file_path.stem
        name = raw_name.strip()
        for possible_suffix in possible_suffixes_to_remove:
            name = name.removesuffix(possible_suffix)
        return name.strip().strip("_")

    def set_changed_parameter(self, changed_parameter: Union["ChangeParameter", str]) -> None:
        if isinstance(changed_parameter, self.ChangeParameter):
            self.changed_parameter = changed_parameter
        else:
            self.changed_parameter = self.ChangeParameter(changed_parameter)

        self.changed_check_func = self.changed_parameter.get_check_method(self)

    def _check_handle_not_existing(self) -> None:
        with self.lock:
            if self.file_path.exists() is True:
                return
            if self.missing_ok is True:
                self.file_path.parent.mkdir(parents=True, exist_ok=True)
                self.file_path.touch(exist_ok=True)
                self._update_changed_data()
                self.file_was_created = True

            else:
                raise FileNotFoundError(f"file for {self.__class__.__name__!r} -> {self.file_path.as_posix()!r} does exist.")

    @ property
    def file_name(self) -> str:
        return self.file_path.name

    @ property
    def size(self) -> int:
        self._check_handle_not_existing()
        size = self.file_path.stat().st_size
        return size

    @ property
    def file_hash(self) -> str:
        self._check_handle_not_existing()
        with self.file_path.open('rb') as f:
            if self.size <= self.file_hash_size_threshold:

                return self.hash_func(f.read()).hexdigest()
            _file_hash = self.hash_func()
            for chunk in f:
                _file_hash.update(chunk)
            return _file_hash.hexdigest()

    @ property
    def mtime(self) -> int:
        self._check_handle_not_existing()
        return self.file_path.stat().st_mtime

    def _on_size_changed_check(self) -> bool:
        return self.last_size is None or self.last_size != self.size

    def _on_file_hash_changed_check(self) -> bool:
        return self.last_file_hash is None or self.last_file_hash != self.file_hash

    def _on_mtime_changed_check(self) -> bool:
        return self.last_mtime is None or self.last_mtime != self.mtime

    def _on_all_changed_check(self) -> bool:
        return any([self._on_size_changed_check(), self._on_file_hash_changed_check(), self._on_mtime_changed_check()])

    def _on_always_changed_check(self) -> bool:
        return True

    def _on_never_changed_check(self) -> bool:
        return False

    @ property
    def has_changed(self) -> bool:
        with self.lock:
            result = self.changed_check_func()
            if result is True:
                self.changed_signal.delayed_fire_and_forget(1, self)
        return result

    def _update_changed_data(self) -> None:

        def _update_size():
            self.last_size = self.size

        def _update_file_hash():
            self.last_file_hash = self.file_hash

        def _update_mtime():
            self.last_mtime = self.mtime

        def _update_all():
            _update_size()
            _update_file_hash()
            _update_mtime()
        update_table = {self.ChangeParameter.NEVER: lambda: ...,
                        self.ChangeParameter.ALWAYS: lambda: ...,
                        self.ChangeParameter.SIZE: _update_size,
                        self.ChangeParameter.FILE_HASH: _update_file_hash,
                        self.ChangeParameter.MTIME: _update_mtime,
                        self.ChangeParameter.ALL: _update_all}
        update_table[self.changed_parameter]()

    @ property
    def _read_kwargs(self) -> dict[str, str]:
        kwargs = {"mode": self.read_mode}
        if 'b' not in self.read_mode:
            kwargs['encoding'] = self._encoding
            kwargs['errors'] = self._on_errors
        return kwargs

    @ property
    def _write_kwargs(self) -> dict[str, str]:
        kwargs = {"mode": self.write_mode}
        if 'b' not in self.write_mode:
            kwargs['encoding'] = self._encoding
            kwargs['errors'] = self._on_errors
        return kwargs

    def read(self):
        self._check_handle_not_existing()
        with self.lock:
            self._update_changed_data()
            # pylint: disable=unspecified-encoding
            with self.file_path.open(**self._read_kwargs) as f:
                return f.read()

    def write(self, data) -> None:
        with self.lock:
            # pylint: disable=unspecified-encoding
            with atomic_write(self.file_path, overwrite=True, **self._write_kwargs) as f:
                f.write(data)

    def __fspath__(self) -> str:
        return str(self.file_path)

    def __str__(self) -> str:
        return self.__fspath__()
# region [Main_Exec]


if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Type
from pathlib import Path

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def implements_protocol(proto: Type):
    """
    Creates a decorator for classes that checks that the decorated class implements the runtime protocol `proto`.

        from https://stackoverflow.com/a/62923698/13989012
    """
    def _deco(cls_def):
        try:
            assert issubclass(cls_def, proto)
        except AssertionError as error:
            error.args = (f"{cls_def} does not implement protocol {proto}",)
            raise
        return cls_def
    return _deco


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
from enum import Flag, auto
from typing import Any, Union, Iterable, Literal, TypeVar
from pathlib import Path
from datetime import timedelta
from operator import neg, or_, pos
from functools import reduce, total_ordering, cached_property
from collections import defaultdict
from pprint import pprint
# * Third Party Imports --------------------------------------------------------------------------------->
import pyparsing as pp
import pyparsing.common as ppc

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import FlagConflictError, UnparsableHumanTimedelta
from gidapptools.data.conversion_data import TimeUnit, TIMEUNITS, TIMEUNIT_NAME_MAP, STRING_TRUE_VALUES, STRING_FALSE_VALUES, FILE_SIZE_SYMBOL_DATA, NANOSECONDS_IN_SECOND, MICROSECONDS_IN_SECOND

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


@total_ordering
class FileSizeUnit:
    __slots__ = ("_short_name", "_long_name", "factor", "aliases", "all_names", "all_names_casefolded")

    def __init__(self, short_name: str, long_name: str, factor: int, aliases: Iterable[str] = None) -> None:
        self._short_name = short_name
        self._long_name = long_name
        self.factor = factor
        self.aliases = [] if aliases is None else list(aliases)
        self.aliases += self._get_default_aliases()
        self.all_names = self._get_names()
        self.all_names_casefolded = {name.casefold() for name in self.all_names}

    @property
    def short_name(self) -> str:
        return f"{self._short_name}b"

    @property
    def long_name(self) -> str:
        return f"{self._long_name}bytes"

    def _get_names(self) -> set[str]:
        all_names: list[str] = [self.short_name, self.long_name]
        all_names += self.aliases
        all_names += [name.removesuffix('s') for name in all_names]
        all_names += [name + 's' for name in all_names if not name.endswith('s')]

        return set(all_names)

    def _get_default_aliases(self) -> Iterable[str]:
        _out = []

        _out.append(f"{self._long_name} bytes")

        return _out

    def __eq__(self, o: object) -> bool:
        if isinstance(o, self.__class__):
            return self.factor == o.factor

        if isinstance(o, int):
            return self.factor == o

        if isinstance(o, str):
            return o in {self.short_name, self.long_name}.union(set(self.aliases))

        return NotImplemented

    def __lt__(self, o: object) -> bool:
        if isinstance(o, self.__class__):
            return self.factor < o.factor
        if isinstance(o, int):
            return self.factor < o

        return NotImplemented

    def __truediv__(self, o: object) -> float:
        if isinstance(o, self.__class__):
            return self.factor / o.factor
        if isinstance(o, int):
            return self.factor / o

        if isinstance(o, float):
            return float(self.factor) / o
        return NotImplemented

    def __rtruediv__(self, o: object) -> float:
        if isinstance(o, self.__class__):
            return o.factor / self.factor
        if isinstance(o, int):
            return o / self.factor

        if isinstance(o, float):
            return o / float(self.factor)

        return NotImplemented

    def __str__(self) -> str:
        return self.short_name


class FileSizeByte(FileSizeUnit):
    __slots__ = ("short_name", "long_name")
    # pylint: disable=super-init-not-called

    def __init__(self) -> None:
        self.short_name = 'b'
        self.long_name = 'bytes'
        self.factor = 1
        self.aliases = []
        self.all_names = self._get_names()
        self.all_names_casefolded = {name.casefold() for name in self.all_names}


class FileSizeReference:
    __slots__ = ("byte_unit", "units")

    def __init__(self) -> None:
        self.byte_unit = FileSizeByte()
        self.units: tuple[FileSizeUnit] = None
        self._make_units()

    def _make_units(self) -> None:
        units: list[FileSizeUnit] = []

        temp_unit_info = {s: 1 << (i + 1) * 10 for i, s in enumerate(FILE_SIZE_SYMBOL_DATA)}
        for key, value in temp_unit_info.items():
            units.append(FileSizeUnit(short_name=key[0], long_name=key[1], factor=value))
        self.units = tuple(sorted(units))

    @property
    def symbols(self) -> tuple[str]:
        return tuple(item.short_name for item in self.units)

    @property
    def long_names(self) -> tuple[str]:
        return tuple(item.long_name for item in self.units)

    def get_unit_by_name(self, name: str, case_insensitive: bool = True) -> FileSizeUnit:
        try:
            all_names = [unit for unit in self.units if name in unit.all_names_casefolded]
            return all_names[0]
        except IndexError as error:
            if name in self.byte_unit.all_names_casefolded:
                return self.byte_unit
            raise KeyError(name) from error


FILE_SIZE_REFERENCE = FileSizeReference()


def bytes2human(n: int) -> str:
    # http://code.activestate.com/recipes/578019

    is_negative = n < 0
    sign_prefix = ""
    if is_negative:
        n = n * (-1)
        sign_prefix = "-"
    for unit in reversed(FILE_SIZE_REFERENCE.units):
        if n >= unit:
            _out = float(n) / unit

            _out = f'{sign_prefix}{_out:.2f} {unit}'
            return _out
    _out = n

    return f"{sign_prefix}{_out} b"


def human2bytes(in_text: str, strict: bool = False) -> int:

    def _clean_name(name: str) -> str:
        name = name.strip()
        name = name.casefold()
        name = white_space_regex.sub(' ', name)
        return name

    if in_text.strip() == "0" and strict is False:
        return ""
    white_space_regex = re.compile(r"\s{2,}")
    number_regex_pattern = r"(?P<number>[\d\.\,]+)"
    name_regex_pattern = r"(?P<name>\w([\w\s]+)?)"
    parse_regex = re.compile(number_regex_pattern + r'\s*' + name_regex_pattern)

    match = parse_regex.match(in_text.strip())
    if match:
        number = float(match.group('number'))
        name = _clean_name(match.group('name'))
        unit = FILE_SIZE_REFERENCE.get_unit_by_name(name)
        return int(number * unit.factor)
    else:
        raise ValueError(f"Unable to parse input string {in_text!r}.")


def ns_to_s(nano_seconds: Union[int, float], decimal_places: int = None) -> Union[int, float]:
    seconds = nano_seconds / NANOSECONDS_IN_SECOND
    if decimal_places is None:
        return seconds
    return round(seconds, decimal_places)


def ms_to_s(micro_seconds: Union[int, float], decimal_places: int = None) -> Union[int, float]:
    seconds = micro_seconds / MICROSECONDS_IN_SECOND
    if decimal_places is None:
        return seconds
    return round(seconds, decimal_places)


class TimeUnits:
    __slots__ = ("_with_year",
                 "units",
                 "name_dict",
                 "symbol_dict",
                 "alias_dict",
                 "full_dict")

    def __init__(self, with_year: bool = True) -> None:
        self._with_year = with_year
        self.units = self._get_units()
        self.name_dict: dict[str, TimeUnit] = {item.name.casefold(): item for item in self.units} | {item.plural.casefold(): item for item in self.units}
        self.symbol_dict: dict[str, TimeUnit] = {item.symbol.casefold(): item for item in self.units}
        self.alias_dict: dict[str, TimeUnit] = self._get_alias_dict()
        self.full_dict: dict[str, TimeUnit] = self.name_dict | self.symbol_dict | self.alias_dict

    @property
    def smallest_unit(self) -> TimeUnit:
        return self.units[-1]

    def _get_units(self):
        _all_units = sorted(TIMEUNITS.copy(), key=lambda x: -x.factor)
        if self._with_year is False:
            return [u for u in _all_units if u.name != 'year']
        return _all_units

    def _get_alias_dict(self) -> dict[str, TimeUnit]:
        _out = {}
        for item in self.units:
            for alias in item.aliases:
                _out[alias] = item
                _out[alias.casefold()] = item
        return _out

    def __getitem__(self, key: Union[int, str]) -> TimeUnit:
        if isinstance(key, int):
            return self.units[key]

        if isinstance(key, str):
            return self.full_dict[key.casefold()]

        return self.full_dict[key]

    def __iter__(self):
        return iter(self.units)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(with_year={self._with_year!r})'


_time_units_without_year = TimeUnits(False)
_time_units_with_year = TimeUnits(True)


def timedelta_to_stopwatch_format(t: Union[float, timedelta]) -> str:
    """
    Get a friendly timestamp represented as a string.
    """
    try:
        all_seconds = t.total_seconds()
    except AttributeError:
        all_seconds = t

    hour_unit = TIMEUNIT_NAME_MAP["hour"]
    minute_unit = TIMEUNIT_NAME_MAP["minute"]
    second_unit = TIMEUNIT_NAME_MAP["second"]

    hours, minutes, seconds = hour_unit.convert_seconds(all_seconds), minute_unit.convert_seconds(all_seconds) % int(hour_unit.factor / minute_unit.factor), all_seconds % int(minute_unit.factor / second_unit.factor)
    return f"{hours:02d}:{minutes:02d}:{seconds:05.2f}"


def seconds2human(in_seconds: Union[int, float, timedelta],
                  as_list_result: bool = False,
                  as_dict_result: bool = False,
                  as_symbols: bool = False,
                  with_year: bool = True,
                  min_unit: str = None) -> Union[dict[TimeUnit, int], str]:
    if as_list_result is True and as_dict_result is True:
        raise FlagConflictError(["as_list_result", "as_dict_result"], True)
    rest = in_seconds.total_seconds() if isinstance(in_seconds, timedelta) else in_seconds
    sign = ""
    if rest < 0:
        rest = abs(rest)
        sign = "-"
    result = {}

    _time_units = _time_units_without_year if with_year is False else _time_units_with_year

    if min_unit is None:
        sub_min_units = set()
    else:
        min_unit = _time_units[min_unit.casefold()]
        sub_min_units = {unit for unit in _time_units if unit.factor < min_unit.factor}

    for unit in _time_units:
        amount, rest = unit.convert_with_rest(rest)
        if amount:
            result[unit] = int(amount)

    results = [k.value_to_string(v, as_symbols) for k, v in result.items() if k not in sub_min_units]
    if as_list_result is True:
        return results
    if as_dict_result is True:
        return {k: v for k, v in result.items() if k not in sub_min_units}

    if not results:
        _unit = _time_units.smallest_unit if min_unit is None else min_unit
        _name = f" {_unit.plural}" if as_symbols is False else _unit.symbol
        return f"0{_name}"
    if len(results) > 1:
        return sign + ' '.join(results[:-1]) + ' and ' + results[-1]
    return sign + results[0]


class TimedeltaConversionModifiers(Flag):
    POSITIVE = auto()
    NEGATIVE = auto()

    @property
    def sign(self):
        if self.__class__.NEGATIVE in self and self.__class__.POSITIVE in self:
            raise ValueError("Parsed timedelta can not have positive and negative modifiers at the same time.")
        if self.__class__.NEGATIVE in self:
            return neg

        return pos


def get_timedelta_parsing_grammar() -> pp.ParserElement:

    possible_time_units = []
    _time_units = TimeUnits(with_year=True)

    def _time_data_action(in_token: pp.ParseResults) -> dict[TimeUnit, int]:
        _out = defaultdict(int)

        for item in in_token:
            key = item[1]
            value = item[0]
            if key == _time_units['y']:
                value = value * key.factor
                key = _time_units['s']
            elif key == _time_units['ns']:
                value = value * key.factor
                key = _time_units['s']

            _out[key] += value

        return dict(_out)

    for unit in _time_units:
        possible_time_units.append(unit.name)
        possible_time_units.append(unit.symbol)
        possible_time_units.append(unit.plural)
        possible_time_units += list(unit.aliases)

    possible_time_units = pp.one_of(possible_time_units, caseless=True).set_parse_action(lambda x: _time_units[x[0]])

    combine_words = pp.one_of(["and", ",", ":", ";"], caseless=True).suppress()
    time_item = pp.Group(ppc.integer + possible_time_units)
    prefixes = pp.Keyword("in", caseless=True).set_parse_action(lambda: TimedeltaConversionModifiers.POSITIVE) | pp.Keyword("since", caseless=True).set_parse_action(
        lambda: TimedeltaConversionModifiers.NEGATIVE) | pp.Literal('-').set_parse_action(lambda: TimedeltaConversionModifiers.NEGATIVE)

    postfixes = pp.Keyword("ago", caseless=True).set_parse_action(lambda: TimedeltaConversionModifiers.NEGATIVE)

    return pp.ZeroOrMore(prefixes)("prefix") + pp.OneOrMore(time_item + pp.Optional(combine_words))('time_data').set_parse_action(_time_data_action) + pp.ZeroOrMore(postfixes)("postfix")


TIMEDELTA_PARSING_GRAMMAR = get_timedelta_parsing_grammar()


def human2timedelta(in_text: str) -> timedelta:

    try:
        tokens = TIMEDELTA_PARSING_GRAMMAR.parse_string(in_text, parse_all=True).as_dict()
    except pp.ParseBaseException as e:
        raise UnparsableHumanTimedelta(in_text) from e
    _raw_modifier_data = tokens.get("prefix") + tokens.get('postfix')
    if _raw_modifier_data == []:
        _raw_modifier_data = [TimedeltaConversionModifiers.POSITIVE]
    modifiers = reduce(or_, {i for i in _raw_modifier_data if i})

    raw_timedelta_kwargs = {k.plural: v for k, v in tokens.get('time_data').items() if v}

    raw_timedelta = timedelta(**raw_timedelta_kwargs)
    return modifiers.sign(raw_timedelta)


def human2seconds(in_text: str) -> float:
    return human2timedelta(in_text=in_text).total_seconds()


def str_to_bool(in_string: str, strict: bool = False) -> bool:
    if isinstance(in_string, bool):
        return in_string
    mod_string = in_string.casefold().strip()
    if strict is False:
        return mod_string in STRING_TRUE_VALUES

    if mod_string in STRING_TRUE_VALUES:
        return True
    if mod_string in STRING_FALSE_VALUES:
        return False

    raise TypeError(f'Unable to convert string {in_string!r} to a Boolean value.')


def number_to_pretty(in_num: Union[int, float]) -> str:

    return f"{in_num:,}"


# region [Main_Exec]

if __name__ == '__main__':
    print(human2bytes("100kb"))
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
from typing import Union, Iterable, Optional, Protocol, TYPE_CHECKING
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
from yarl import URL

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtWidgets import QApplication

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_utility.version_item import VersionItem
import sys
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
if TYPE_CHECKING:
    from gidapptools.meta_data.meta_info.meta_info_item import MetaInfo

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class ApplicationInfo(Protocol):
    app_name: str
    app_author: str
    version: Optional[Union[str, VersionItem]]
    url: Optional[Union[str, URL]]


class GidBaseApplication(QApplication):

    def __init__(self, argv: Iterable[str] = None):
        super().__init__(argv or sys.argv)
        self.is_setup: bool = False

    @classmethod
    def is_ready(cls) -> bool:
        return cls.startingUp() is False and cls.instance().is_setup is True

    def setup_from_meta_info(self,
                             meta_info: "MetaInfo") -> Self:
        ...

    def setup(self,
              app_name: str = None,
              app_author: str = None,
              version: Union[str, "VersionItem"] = None,
              url: Union[URL, str] = None) -> Self:
        ...
# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import re
import sys
import queue
import atexit
import logging
import traceback
import warnings
from typing import TYPE_CHECKING, Any, Union, Mapping, Callable, Iterable, Optional, TextIO, TypeAlias
from pathlib import Path
from types import ModuleType
from weakref import WeakValueDictionary
from contextlib import contextmanager
from logging.handlers import QueueHandler, QueueListener
import importlib
# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_logger.enums import LoggingLevel
from gidapptools.gid_logger.handler import GidBaseStreamHandler, GidBaseRotatingFileHandler
from gidapptools.gid_logger.formatter import GidLoggingFormatter, get_all_func_names, get_all_module_names
from gidapptools.general_helper.conversion import str_to_bool

import sys
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gid_logger.records import LOG_RECORD_TYPES

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def is_dev() -> bool:
    is_dev_string = os.getenv("IS_DEV", '0').casefold()
    return str_to_bool(is_dev_string, strict=True) or sys.flags.dev_mode


class GidLogger(logging.Logger):
    main_pack_name: str = None

    def __init__(self, name: str, level: "logging._Level" = logging.NOTSET) -> None:
        super().__init__(name, level)
        self.que_listener: QueueListener = None
        self._que_handlers: WeakValueDictionary[str, logging.Handler] = WeakValueDictionary()

    @property
    def all_handlers(self) -> dict[str, tuple[logging.Handler]]:
        return {"handlers": tuple(self.handlers),
                "que_handlers": dict(self._que_handlers)}

    def set_que_listener(self, que_listener: QueueListener):
        def _determine_handler_name(in_handler: logging.Handler) -> str:
            if hasattr(in_handler, "name") and in_handler.name is not None:
                return in_handler.name
            return in_handler.__class__.__name__

        self.que_listener = que_listener
        for handler in que_listener.handlers:
            self._que_handlers[_determine_handler_name(handler)] = handler

    def makeRecord(self,
                   name: str,
                   level: int,
                   fn: str,
                   lno: int,
                   msg: object,
                   args: "logging._ArgsType",
                   exc_info: "logging._SysExcInfoType" = None,
                   func: str = None,
                   extra: Mapping[str, object] = None,
                   sinfo: str = None) -> "LOG_RECORD_TYPES":
        rv = super().makeRecord(name, level, fn, lno, msg, args=args, exc_info=exc_info, func=func, extra=extra, sinfo=sinfo)
        if not hasattr(rv, "extras"):
            setattr(rv, "extras", {})
        if extra is not None:
            rv.extras |= extra

        return rv

    def _log(self, level: int, msg: object, args: "logging._ArgsType", exc_info: "logging._ExcInfoType" = None, extra: Mapping[str, object] = None, stack_info: bool = False, stacklevel: int = 2) -> None:
        if extra is not None and extra.get("is_timing_decorator", False) is True:
            stacklevel += 0
        return super()._log(level, msg, args, exc_info, extra, stack_info, stacklevel)


class MetaLogger(GidLogger):

    def _log(self, level: int, msg: object, args: "logging._ArgsType", exc_info: "logging._ExcInfoType" = None, extra: Mapping[str, object] = None, stack_info: bool = False, stacklevel: int = 2) -> None:
        extra = extra or {}
        extra["is_meta"] = True

        return super()._log(level, msg, args, exc_info, extra, stack_info, stacklevel)


def make_library_logger(in_name: str) -> logging.Logger:
    logger = logging.getLogger(in_name)
    logger.addHandler(logging.NullHandler())
    return logger


@contextmanager
def switch_logger_klass(logger_klass: type[logging.Logger]):
    original_logger_klass = logging.getLoggerClass()
    try:
        logging.setLoggerClass(logger_klass)
        yield
    finally:
        logging.setLoggerClass(original_logger_klass)


def _modify_logger_name(name: str) -> str:
    if name == "__logging_meta__":
        return 'main.__logging_meta__'
    # name = 'main.' + '.'.join(name.split('.')[1:])
    if name.startswith("__main__"):
        return name.replace("__main__", "main")

    if GidLogger.main_pack_name is not None:

        return name.replace(GidLogger.main_pack_name, "main")

    return name


def get_main_logger() -> logging.Logger:
    return get_logger("__main__")


META_LOGGER_NAME = "__logging_meta__"


def get_meta_logger():
    main_logger = get_main_logger()
    with switch_logger_klass(MetaLogger):
        return main_logger.getChild(META_LOGGER_NAME)


def get_logger(name: str) -> Union[logging.Logger, GidLogger]:
    name = _modify_logger_name(name)
    with switch_logger_klass(GidLogger):
        return logging.getLogger(name)


def get_handlers(logger: Union[logging.Logger, GidLogger] = None) -> tuple[logging.Handler]:
    logger = logger or get_main_logger()
    handlers = logger.handlers
    all_handlers = []
    for handler in handlers:
        all_handlers.append(handler)
    return tuple(all_handlers)


class WarningItem:
    __slots__ = ("_message",
                 "category",
                 "_raw_filename",
                 "lineno",
                 "file",
                 "line",
                 "module",
                 "resolved_filename",
                 "func_name",
                 "_is_resolved")

    def __init__(self,
                 message: str,
                 category: type[Warning],
                 filename: str,
                 lineno: int,
                 file: Optional[TextIO] = None,
                 line: str = None) -> None:
        self._message = message
        self.category = category
        self._raw_filename = filename
        self.lineno = lineno
        self.file = file
        self.line = line

        self.module: Optional[ModuleType] = None
        self.resolved_filename: Optional[Path] = None
        self.func_name: Optional[str] = None
        self._is_resolved: bool = False

    @property
    def message(self) -> str:
        return str(self._message).rstrip("\n")

    @property
    def is_frozen_stdlib(self) -> bool:
        return self._raw_filename.startswith("<frozen ") and self._raw_filename.endswith(">")

    @property
    def filename(self) -> str:
        if self.resolved_filename is None:
            return self._raw_filename
        return str(self.resolved_filename)

    @property
    def file_path(self) -> Optional[Path]:
        if self.resolved_filename is not None:
            return self.resolved_filename
        return None

    @property
    def module_name(self) -> str:
        if self.module is not None:
            return self.module.__name__

        if self.is_frozen_stdlib is True:
            return self.filename.removeprefix("<frozen ").removesuffix(">").strip()

        return self.filename

    def _resolve_filename(self) -> None:
        if self._raw_filename.startswith("<frozen ") and self._raw_filename.endswith(">"):
            package_name, *sub_names = self._raw_filename.removeprefix("<frozen ").removesuffix(">").strip().rsplit(".", 1)
            package = sys.modules[package_name]
            module = package
            for sub_name in sub_names:
                module = getattr(module, sub_name)
            self.module = module
            self.resolved_filename = Path(module.__file__).resolve()

        elif os.path.isfile(self._raw_filename) is True:
            self.resolved_filename = Path(self._raw_filename).resolve()
            for _module in tuple(sys.modules.values()):
                if Path(_module.__name__).resolve() == self.resolved_filename:
                    self.module = _module
                    break

    def _resolve_line(self) -> None:
        if self.resolved_filename is None:
            return
        import linecache
        temp_lineno = self.lineno
        line = linecache.getline(str(self.resolved_filename), temp_lineno)
        if self.line is None:
            self.line = line.rstrip()

        while True:
            if match := func_pattern.search(line):
                self.func_name = match.group("func_name").strip()
                break
            temp_lineno -= 1
            if temp_lineno < 0:
                self.func_name = "module"
                break
            line = linecache.getline(str(self.resolved_filename), temp_lineno)

    def resolve(self) -> Self:
        if self._is_resolved is True:
            return
        self._resolve_filename()
        self._resolve_line()
        self._is_resolved = True
        return self

    def get_logging_params(self) -> dict[str, object]:
        _out = {}
        _out["name"] = self.module_name
        _out["level"] = logging.WARNING
        _out["fn"] = self.func_name
        _out["lno"] = self.lineno
        _out["msg"] = self.message
        _out["args"] = tuple()
        _out["exc_info"] = None
        _out["func"] = self.func_name

        return _out

    def __repr__(self) -> str:
        _out = self.__class__.__name__ + "("

        for attr_name in [i for i in self.__slots__ if not i.startswith("_")] + ["message", "filename", "is_frozen_stdlib"]:
            _out += f"{attr_name}={getattr(self, attr_name)!r}, "

        _out.removesuffix(", ")
        _out += ")"
        return _out

    def __str__(self) -> str:
        return self.message


func_pattern = re.compile(r"(^| )def (?P<func_name>[^\(\)\n\"\'\]\[]+)")


SHOW_WARNING_FUNC_TYPE: TypeAlias = Callable[[str, type[Warning], str, int, Optional[str], Optional[str]], None]


class WarningHandler:
    __slots__ = ("_old_show_warnings_func",)

    def __init__(self, old_show_warnings_func: Optional[SHOW_WARNING_FUNC_TYPE] = None) -> None:
        self._old_show_warnings_func = old_show_warnings_func or self._fallback_show_warning

    @property
    def meta_logger(self) -> logging.Logger:
        return get_meta_logger()

    def _fallback_show_warning(self, message: str, category: type[Warning], filename: str, lineno: int, file: Optional[TextIO] = None, line: Optional[str] = None) -> None:
        self.meta_logger.warning("%s - %s:%s:%r - %s", category.__name__, filename, lineno, line or "", message)

    def _show_warnings(self, message: str, category: type[Warning], filename: str, lineno: int, file: Optional[TextIO] = None, line: Optional[str] = None) -> None:
        try:
            warning_item = WarningItem(message=message, category=category, filename=filename, lineno=lineno, file=file, line=line).resolve()
            logger = get_logger(f"__main__.{warning_item.module_name}")
            record = logger.makeRecord(**warning_item.get_logging_params())
            logger.handle(record)
        except Exception as e:
            self.meta_logger.error(e, exc_info=True)
            self._old_show_warnings_func(message=message, category=category, filename=filename, lineno=lineno, line=line)

    def __call__(self, message: str, category: type[Warning], filename: str, lineno: int, file: Optional[TextIO] = None, line: str = None) -> None:
        self._show_warnings(message=message, category=category, filename=filename, lineno=lineno, file=file, line=line)


def setup_main_logger(name: str,
                      path: Path,
                      log_level: LoggingLevel = LoggingLevel.DEBUG,
                      formatter: Union[logging.Formatter, GidLoggingFormatter] = None,
                      extra_logger: Iterable[str] = tuple(),
                      *,
                      determine_max_module_len: bool = False,
                      determine_max_func_name_len: bool = False) -> Union[logging.Logger, GidLogger]:
    # if determine_max_func_name_len:
    #     os.environ["MAX_FUNC_NAME_LEN"] = str(min([max(len(i) for i in get_all_func_names(path, True)), 20]))
    # if determine_max_module_len:
    #     os.environ["MAX_MODULE_NAME_LEN"] = str(min([max(len(i) for i in get_all_module_names(path)), 20]))

    handler = GidBaseStreamHandler(stream=sys.stdout)

    que = queue.Queue(-1)
    que_handler = QueueHandler(que)
    listener = QueueListener(que, handler)
    formatter = GidLoggingFormatter() if formatter is None else formatter
    handler.setFormatter(formatter)
    _log = get_logger(name)
    for logger in [_log] + [logging.getLogger(l) for l in extra_logger]:
        logger.addHandler(que_handler)

        logger.setLevel(log_level)
    _log.addHandler(que_handler)
    _log.setLevel(log_level)
    listener.start()
    atexit.register(listener.stop)
    return _log


def setup_main_logger_with_file_logging(name: str,
                                        log_file_base_name: str,
                                        path: Path,
                                        log_level: LoggingLevel = LoggingLevel.DEBUG,
                                        formatter: Union[logging.Formatter, GidLoggingFormatter] = None,
                                        log_folder: Path = None,
                                        extra_logger: Iterable[str] = tuple(),
                                        max_func_name_length: int = None,
                                        max_module_name_length: int = None,
                                        *,
                                        log_to_file: bool = True,
                                        log_to_stdout: bool = True,
                                        main_pack_name: str = None) -> Union[logging.Logger, GidLogger]:
    GidLogger.main_pack_name = main_pack_name
    if is_dev() is True:
        log_folder = path.parent.joinpath('logs')

    # os.environ["MAX_FUNC_NAME_LEN"] = str(max_func_name_length) if max_func_name_length is not None else str(min([max(len(i) for i in get_all_func_names(path, True)), 25]))
    # os.environ["MAX_MODULE_NAME_LEN"] = str(max_module_name_length) if max_module_name_length is not None else str(min([max(len(i) for i in get_all_module_names(path)), 25]))

    que = queue.Queue()
    que_handler = QueueHandler(que)

    formatter = GidLoggingFormatter() if formatter is None else formatter
    endpoints = []
    if log_to_stdout is True:
        handler = GidBaseStreamHandler()
        handler.setFormatter(formatter)
        endpoints.append(handler)
    if log_to_file is True:
        file_handler = GidBaseRotatingFileHandler(base_name=log_file_base_name, log_folder=log_folder)

        file_handler.setFormatter(formatter)
        endpoints.append(file_handler)
    # storing_handler = GidStoringHandler(50)
    # storing_handler.setFormatter(formatter)
    # endpoints.append(storing_handler)

    listener = QueueListener(que, *endpoints)
    _log = get_logger(name)
    log_level = LoggingLevel(log_level)
    if "py.warnings" in extra_logger:
        logging.captureWarnings(True)
        warning_handler = WarningHandler(warnings.showwarning)
        warnings.showwarning = warning_handler
    for logger in [_log] + [logging.getLogger(l) for l in extra_logger]:
        logger.addHandler(que_handler)

        logger.setLevel(log_level)
    listener.start()
    atexit.register(listener.stop)
    _log.set_que_listener(listener)
    return _log


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING
from pathlib import Path
from enum import Enum, auto
# * Qt Imports --------------------------------------------------------------------------------------->
import PySide6
from PySide6.QtCore import Qt, QEvent, QSize
from PySide6.QtWidgets import QWidget, QGroupBox, QGridLayout, QSizePolicy

from PySide6 import (QtCore, QtGui, QtWidgets, Qt3DAnimation, Qt3DCore, Qt3DExtras, Qt3DInput, Qt3DLogic, Qt3DRender, QtAxContainer, QtBluetooth,
                     QtCharts, QtConcurrent, QtDataVisualization, QtDesigner, QtHelp, QtMultimedia, QtMultimediaWidgets, QtNetwork, QtNetworkAuth,
                     QtOpenGL, QtOpenGLWidgets, QtPositioning, QtPrintSupport, QtQml, QtQuick, QtQuickControls2, QtQuickWidgets, QtRemoteObjects,
                     QtScxml, QtSensors, QtSerialPort, QtSql, QtStateMachine, QtSvg, QtSvgWidgets, QtTest, QtUiTools, QtWebChannel, QtWebEngineCore,
                     QtWebEngineQuick, QtWebEngineWidgets, QtWebSockets, QtXml)

from PySide6.QtCore import (QByteArray, QCoreApplication, QDate, QDateTime, QEvent, QLocale, QMetaObject, QModelIndex, QModelRoleData, QMutex,
                            QMutexLocker, QObject, QPoint, QRect, QRecursiveMutex, QRunnable, QSettings, QSize, QThread, QThreadPool, QTime, QUrl,
                            QWaitCondition, Qt, QAbstractItemModel, QAbstractListModel, QAbstractTableModel, Signal, Slot, QMargins)

from PySide6.QtGui import (QAction, QBrush, QColor, QPaintEvent, QConicalGradient, QCursor, QFont, QFontDatabase, QFontMetrics, QGradient, QIcon, QImage,
                           QKeySequence, QLinearGradient, QPainter, QPalette, QPixmap, QRadialGradient, QTransform)

from PySide6.QtWidgets import (QApplication, QBoxLayout, QCheckBox, QColorDialog, QColumnView, QComboBox, QDateTimeEdit, QDialogButtonBox,
                               QDockWidget, QDoubleSpinBox, QFontComboBox, QFormLayout, QFrame, QGridLayout, QGroupBox, QHBoxLayout, QHeaderView,
                               QLCDNumber, QLabel, QLayout, QLineEdit, QListView, QListWidget, QMainWindow, QMenu, QMenuBar, QMessageBox,
                               QProgressBar, QProgressDialog, QPushButton, QSizePolicy, QSpacerItem, QSpinBox, QStackedLayout, QStackedWidget,
                               QStatusBar, QStyledItemDelegate, QSystemTrayIcon, QTabWidget, QTableView, QTextEdit, QTimeEdit, QToolBox, QTreeView,
                               QVBoxLayout, QWidget, QAbstractItemDelegate, QAbstractItemView, QAbstractScrollArea, QRadioButton, QFileDialog, QButtonGroup, QWidgetItem, QLayoutItem)

# * Gid Imports ----------------------------------------------------------------------------------------->


# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    pass

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]


THIS_FILE_DIR = Path(__file__).parent.absolute()


# endregion [Constants]


class EmptyPlaceholderWidget(QWidget):

    @property
    def has_content(self) -> bool:
        return False


class CollapsibleGroupBox(QGroupBox):
    _expand_prefix = "▼"
    _collapse_prefix = "▲"
    _expand_text = "show"
    _collapse_text = "hide"
    _no_content_text = ""
    _no_content_prefix = ""

    def __init__(self,
                 text: str = None,
                 content: QWidget = None,
                 start_expanded: bool = True,
                 parent=None):
        super().__init__(parent=parent)
        self.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)
        self.setLayout(QGridLayout())
        self.layout.setAlignment(Qt.AlignCenter)
        self.setFlat(True)
        self.setAlignment(Qt.AlignCenter)
        self.no_content_prefix = str(self._no_content_prefix)
        self.no_content_text = str(self._no_content_text)
        self.expand_prefix = str(self._expand_prefix)
        self.collapse_prefix = str(self._collapse_prefix)
        self.expand_text = str(self._expand_text)
        self.collapse_text = str(self._collapse_text)
        self.text = text
        self.content = content or EmptyPlaceholderWidget()
        self.layout.addWidget(self.content)
        self.expanded = True
        self.setTitle(self.full_text)
        self.original_cursor = self.cursor()
        self.current_cursor = self.cursor()
        if start_expanded is False:
            self.set_expanded(False)

    @property
    def has_content(self) -> bool:
        return getattr(self.content, "has_content", self.content is not None)

    @property
    def full_text(self) -> str:

        if self.has_content is False:
            text = self.no_content_text
            prefix = self.no_content_prefix

        elif self.expanded is True:
            text = self.text or self.collapse_text
            prefix = self.collapse_prefix + " "

        else:
            text = self.text or self.expand_prefix
            prefix = self.expand_prefix + " "
        return f"{prefix}{text}"

    @ property
    def layout(self) -> QGridLayout:
        return super().layout()

    def mousePressEvent(self, event: PySide6.QtGui.QMouseEvent) -> None:
        if not self.contentsRect().contains(event.position().toPoint()):
            self.on_title_clicked()

    def event(self, event: PySide6.QtCore.QEvent) -> bool:
        if event.type() == QEvent.HoverMove:
            if self.has_content is True and not self.contentsRect().contains(event.position().toPoint()) and self.current_cursor is not Qt.PointingHandCursor:
                self.setCursor(Qt.PointingHandCursor)
            elif self.contentsRect().contains(event.position().toPoint()) and self.current_cursor is not self.original_cursor:
                self.setCursor(self.original_cursor)

        return super().event(event)

    def leaveEvent(self, event: PySide6.QtCore.QEvent) -> None:
        self.setCursor(self.original_cursor)
        return super().leaveEvent(event)

    def on_title_clicked(self):
        if self.has_content:
            self.set_expanded(not self.expanded)

    def set_expanded(self, value: bool):
        self.expanded = value
        self.content.setVisible(value)
        self.setTitle(self.full_text)
        self.update()

    def set_content(self, content: QWidget) -> None:
        if self.has_content is True:
            self.layout.removeWidget(self.content)
        self.content = content
        self.layout.addWidget(self.content)
        self.setTitle(self.full_text)

        if self.has_content is False and self.expanded is True:
            self.set_expanded(False)

    def sizeHint(self) -> QSize:
        # return super().sizeHint()
        if self.expanded is False:
            return super().sizeHint() + self.content.sizeHint()

        else:
            return super().sizeHint()


# region [Main_Exec]


if __name__ == '__main__':
    from gidapptools.gidapptools_qt.helper.misc import batch_modify_widget
    x = QApplication()
    w = QMainWindow()
    cont_wid = QTextEdit()
    batch_modify_widget(cont_wid, setFontItalic=True, setFontPointSize=(10.5,), paste=...)
    # cont_wid.setText("sadfsdfgsfsedfsefdsdfsdf\nasdfasdasdasdasda\nsadfsdfsdfsdfsdf")
    wid = CollapsibleGroupBox(text="wuff", content=cont_wid)
    w.setCentralWidget(wid)
    w.show()
    x.exec()

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from enum import Enum, auto
from typing import TYPE_CHECKING, Union, Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6 import QtGui
from PySide6.QtGui import QPixmap
from PySide6.QtCore import Qt, QSize, Signal
from PySide6.QtWidgets import QFrame, QLabel, QWidget, QGroupBox, QGridLayout, QHBoxLayout, QSizePolicy, QVBoxLayout, QStackedWidget, QDialogButtonBox

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.string_helper import StringCase, StringCaseConverter

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    ...

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class SelectorPosition(Enum):
    TOP = auto()
    Bottom = auto()
    LEFT = auto()
    RIGHT = auto()


class CategoryPicture(QFrame):
    clicked = Signal(int)

    def __init__(self, text: str, picture: Optional[QPixmap], category_page_number: int, parent=None) -> None:
        super().__init__(parent=parent)
        self.raw_text = text
        self.raw_picture = picture
        self.text: QLabel = None
        self.picture: QLabel = None
        self.category_page_number = category_page_number
        self.base_style = QFrame.Raised | QFrame.Panel

    def setup(self) -> "CategoryPicture":
        self.setLayout(QVBoxLayout(self))
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.setContentsMargins(0, 0, 0, 1)
        self.setFrameStyle(self.base_style)
        self.setMidLineWidth(3)
        self.setLineWidth(3)

        self.setup_text()
        self.setup_picture()
        return self

    def setup_text(self):
        self.setToolTip(self.raw_text)
        self.text = QLabel(text=self.raw_text, parent=self)
        self.text.setAlignment(Qt.AlignHCenter | Qt.AlignBottom)
        self.text.setTextFormat(Qt.AutoText)

        self.layout.addWidget(self.text)

    def setup_picture(self):
        self.picture = QLabel(self)
        if self.raw_picture:
            picture = self.raw_picture.scaled(60, 60, Qt.KeepAspectRatioByExpanding, Qt.SmoothTransformation)
            self.picture.setPixmap(picture)
        self.picture.setAlignment(Qt.AlignCenter)

        self.layout.addWidget(self.picture)

    def sizeHint(self) -> QSize:
        width = self.text.sizeHint().width() + self.picture.sizeHint().width()
        height = self.text.sizeHint().height() + self.picture.sizeHint().height()
        return QSize(w=width, h=height)

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        self.setFrameStyle(QFrame.Sunken | QFrame.Panel)


    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        self.setFrameStyle(self.base_style)
        self.clicked.emit(self.category_page_number)

    @property
    def layout(self) -> QVBoxLayout:
        return super().layout()

    def __repr__(self) -> str:
        """
        Basic Repr
        !REPLACE!
        """
        return f'{self.__class__.__name__}'


class PictureCategorySelector(QGroupBox):

    def __init__(self, content_widget: QStackedWidget, parent: QStackedWidget = None, position: SelectorPosition = SelectorPosition.LEFT):
        super().__init__(parent=parent)
        self.position = position
        self.content_widget = content_widget
        self._set_layout()
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.MinimumExpanding)

        self.layout.setAlignment(Qt.AlignCenter)
        self.setFlat(True)
        # self.setTitle("Categories")
        self.categories: dict[str, CategoryPicture] = {}

    def _set_layout(self):
        if self.position in {SelectorPosition.TOP, SelectorPosition.Bottom}:
            layout_class = QHBoxLayout
        elif self.position in {SelectorPosition.LEFT, SelectorPosition.RIGHT}:
            layout_class = QVBoxLayout

        self.setLayout(layout_class(self))

    @property
    def layout(self) -> Union[QHBoxLayout, QVBoxLayout]:
        return super().layout()

    def add_category(self, name: str, picture: Optional[QPixmap], category_page_number: int, verbose_name: str = None):
        verbose_name = verbose_name or StringCaseConverter.convert_to(name, StringCase.SPLIT, clean_in_string=True).title()

        category = CategoryPicture(verbose_name, picture, category_page_number, self).setup()

        self.layout.addWidget(category)
        self.categories[name] = category

        category.clicked.connect(self.content_widget.setCurrentIndex)
        self.resize_categories()

    def sizeHint(self) -> QSize:
        width = max(c.sizeHint().width() for c in self.categories.values())
        height = sum(c.height() for c in self.categories.values())
        return QSize(w=width, h=int(height * 1.25))

    def resize_categories(self):
        widths = []
        heights = []
        for category in self.categories.values():
            widths.append(category.sizeHint().width())
            heights.append(category.height())

        max_width = max(widths)
        max_heights = max(heights)
        for category in self.categories.values():
            category.resize(max_width, max_heights)
        self.repaint()

    def __repr__(self) -> str:
        """
        Basic Repr
        !REPLACE!
        """
        return f'{self.__class__.__name__}'


class ContentStackedwidget(QStackedWidget):

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent=parent)
        self.pages: dict[str, QWidget] = {}
        self.setFrameShape(QFrame.WinPanel)
        self.setFrameShadow(QFrame.Sunken)
        self.setLineWidth(2)
        self.setMidLineWidth(2)

    def addWidget(self, w: QWidget, name: str = None) -> int:
        name = name or w.category_name
        self.pages[name] = w
        return super().addWidget(w)

    def __repr__(self) -> str:
        """
        Basic Repr
        !REPLACE!
        """
        return f'{self.__class__.__name__}'


class CategorySelectWidget(QWidget):
    _available_selector_positions = {SelectorPosition.TOP, SelectorPosition.LEFT}

    def __init__(self, selector_position: SelectorPosition = SelectorPosition.LEFT, parent=None) -> None:
        super().__init__(parent=parent, f=Qt.Dialog)
        self.selector_position = selector_position
        if self.selector_position not in self._available_selector_positions:
            raise ValueError(f"Selector Posiotion {self.selector_position!r} is not available, only {', '.join(repr(i) for i in self._available_selector_positions)}")
        self.main_layout: QGridLayout = None
        self.buttons: QDialogButtonBox = None
        self.selection_box: PictureCategorySelector = None
        self.content_widget: ContentStackedwidget = None

    def setup(self) -> "CategorySelectWidget":

        self.main_layout = QGridLayout()
        self.setLayout(self.main_layout)

        self.setup_buttons()
        self.setup_content_widget()
        self.setup_selection_box()
        return self

    def setup_buttons(self) -> None:
        self.buttons = QDialogButtonBox(self)
        self.buttons.setOrientation(Qt.Horizontal)
        self.buttons.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        if self.selector_position is SelectorPosition.TOP:
            self.main_layout.addWidget(self.buttons, 2, 0, 1, 1, Qt.AlignBottom)
        elif self.selector_position is SelectorPosition.LEFT:
            self.main_layout.addWidget(self.buttons, 1, 1, 1, 1, Qt.AlignBottom)
        self.buttons.rejected.connect(self.on_cancelled)
        self.buttons.accepted.connect(self.on_accepted)

    def setup_content_widget(self) -> None:
        self.content_widget = ContentStackedwidget(self)
        if self.selector_position is SelectorPosition.TOP:
            self.main_layout.addWidget(self.content_widget, 1, 0, 1, 1)
        elif self.selector_position is SelectorPosition.LEFT:
            self.main_layout.addWidget(self.content_widget, 0, 1, 1, 1)

    def setup_selection_box(self) -> None:
        self.selection_box = PictureCategorySelector(self.content_widget, position=self.selector_position)
        if self.selector_position is SelectorPosition.TOP:
            self.main_layout.addWidget(self.selection_box, 0, 0, 1, 1, Qt.AlignTop)
        elif self.selector_position is SelectorPosition.LEFT:
            self.main_layout.addWidget(self.selection_box, 0, 0, 1, 1, Qt.AlignLeft | Qt.AlignTop)

    def add_category(self, content_widget: QWidget):
        try:
            name = content_widget.name
        except AttributeError:
            name = content_widget.objectName()
            if not name:
                name = str(content_widget)

        try:
            picture = content_widget.category_picture
        except AttributeError:
            picture = None
        page_number = self.content_widget.addWidget(content_widget, name=name)
        self.selection_box.add_category(name, picture, page_number)

    def change_page(self, current_item, previous_item):
        self.content_widget.setCurrentIndex(current_item.page_number)

    def on_cancelled(self):
        self.close()

    def on_accepted(self):
        self.close()

    def __repr__(self) -> str:
        """
        Basic Repr
        !REPLACE!
        """
        return f'{self.__class__.__name__}'

# region [Main_Exec]


if __name__ == '__main__':
    # app = QApplication()

    # select_widget = CategorySelectWidget(selector_position=SelectorPosition.LEFT).setup()
    # content = QTextEdit()
    # content.name = "first"
    # content.category_picture = QPixmap(r"D:\Dropbox\hobby\Modding\Programs\Github\My_Repos\GidAppTools\docs\source\_images\broca_progs_logo.png")
    # content.setText("osdfolijsdolpfjlsdfjlksdjflksdjfljsd psdjfpöosdjfpdsfj \n\n sdfosidfjlpsdjfoljsdflsdf")
    # content2 = QDoubleSpinBox()
    # content2.name = "second"

    # content2.category_picture = QPixmap(r"D:\Dropbox\hobby\Modding\Ressources\Icons\To_Sort_Icons\png_icons\document(1).png")
    # select_widget.add_category(content_widget=content)
    # select_widget.add_category(content_widget=content2)
    # select_widget.show()
    # app.exec()
    pass

# endregion [Main_Exec]
